"""
Classes for allowing reload of database.

This isn't meant to cover all possible use cases,
but the conveniently allow restarting from a checkpoint
file without manually re-loading the database.

Restarting a database should only be performed
if no preprocessing is applied, i.e. in order
for this functionality to work automatically, you should
save a copy of your database completely preprocessed.
Otherwise you'll need to reproduce the processing
that was applied between constructing the database object
and constructing the checkpoint.
"""


class Restarter:
    def attempt_reload(self):
        return NotImplemented


class NoRestart(Restarter):
    def attempt_reload(self):
        print("Couldn't reload database. It might have been generated in-memory.")
        return None


class RestartDB(Restarter):
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def __getstate__(self):
        state = self.__dict__.copy()
        cls_spec = (self.cls.__module__, self.cls.__qualname__)
        state["cls"] = cls_spec
        return state

    def __setstate__(self, state):
        cls_module, cls_name = state["cls"]
        try:
            import importlib

            module = importlib.import_module(cls_module)
            state["cls"] = getattr(module, cls_name)
        except (ImportError, AttributeError) as ee:
            # Save the message, not the full exception object
            # including traceback etc.
            state["cls"] = ee.msg
        for k, v in state.items():
            setattr(self, k, v)

    def attempt_reload(self):
        print("restarting", self.cls)
        if isinstance(self.cls, str):
            raise RuntimeError(f"Not restartable due to error: {self.cls}")
        try:
            db = self.cls(*self.args, **self.kwargs)
        except Exception as eee:
            raise RuntimeError("Exception while attempting to reload the database.") from eee
        return db


class Restartable:
    @classmethod
    def make_restarter(cls, *args, **kwargs):
        return RestartDB(cls, *args, **kwargs)



"""
Interface to use schnetpack networks with hippynn.

Note: Only open boundary conditions are supported for now.

"""
try:
    import schnetpack
except ImportError as ie:
    raise ImportError("Schnetpack installation is required to use the schnetpack interface.") from ie

import torch

from schnetpack import Properties

from ...graphs.nodes.base import SingleNode, AutoKw
from ...graphs.nodes.networks import Network
from ...graphs.indextypes import IdxType


class SchNetWrapper(torch.nn.Module):
    def __init__(self, schnet):
        super().__init__()
        self.schnet = schnet
        feature_sizes = [x.dense.out_features for x in schnet.interactions]
        if not schnet.return_intermediates:
            feature_sizes = [feature_sizes[-1]]
        self.feature_sizes = feature_sizes

    def forward(self, z_arr, r_arr, nonblank):
        """
        Wrap a call into the underlying schnet, which uses
        dictionaries.
        :param z_arr:
        :param r_arr:
        :param nonblank:
        :return:
        """
        packed = create_schnetpack_inputs(z_arr, r_arr, nonblank)
        outputs = self.schnet(packed)
        if self.schnet.return_intermediate:
            outputs = outputs[1]
        else:
            outputs = [outputs]
        outputs = [x[packed["nonblank"]] for x in outputs]
        return outputs


class SchNetNode(AutoKw, Network, SingleNode):
    _input_names = "species", "positions", "nonblank"
    _index_state = IdxType.Atoms
    _auto_module_class = SchNetWrapper

    def __init__(self, name, parents, module="auto", module_kwargs=None):
        if module == "auto":
            self.module_kwargs = module_kwargs
            module = self.auto_module()
        super().__init__(name, parents, module=module)


def create_schnetpack_inputs(z_arr, r_arr, nonblank):

    dtype = r_arr.dtype
    device = r_arr.device
    n_atoms_per_mol = (z_arr > 0).sum(axis=1)

    n_mols = n_atoms_per_mol.shape[0]
    n_atoms = n_atoms_per_mol.max()
    z_arr = z_arr[:, :n_atoms]
    r_arr = r_arr[:, :n_atoms]
    nonblank = nonblank[:, :n_atoms]

    atom_range = torch.arange(n_atoms).unsqueeze(0).expand(n_mols, -1)

    atom_mask = torch.zeros(n_mols, n_atoms, dtype=z_arr.dtype, device=device)
    atom_mask[atom_range < n_atoms_per_mol.unsqueeze(1)] = 1

    neighbor_base = torch.arange(n_atoms - 1, device=device, dtype=dtype).unsqueeze(0).expand(n_mols, n_atoms, -1)
    neighbor_base = neighbor_base + torch.triu(torch.ones_like(neighbor_base), diagonal=0)

    neighbor_mask = atom_mask.unsqueeze(2) * atom_mask.unsqueeze(1)[:, :, 1:]
    neighbors = neighbor_base * neighbor_mask

    cell = torch.zeros((n_mols, 3, 3), device=device, dtype=dtype)
    cell_offset = torch.zeros((n_mols, n_atoms, n_atoms - 1, 3), device=device, dtype=dtype)

    return {
        Properties.atom_mask: atom_mask.to(dtype),
        Properties.neighbors: neighbors,
        Properties.neighbor_mask: neighbor_mask.to(dtype),
        Properties.Z: z_arr,
        Properties.R: r_arr,
        Properties.cell: cell,
        Properties.cell_offset: cell_offset,
        "nonblank": nonblank,
    }



"""
Evaluating network performance over a dataset
"""
import torch
from hippynn.tools import device_fallback, progress_bar


class Evaluator:
    def __init__(self, model, evaluation_loss, evaluation_loss_names, plot_maker=None, plot_every=1, db_info=None):
        """

        :param model: ``GraphModule``. The model to evaluate .
        :param evaluation_loss: ``GraphModule``. The losses to evaluate.
        :param evaluation_loss_names: The names of the losses in the evaluation_loss
        :param plot_maker: An optional plot maker to run.
        :param plot_every: How often to make plots.
        :param db_info:
        """

        self.model = model
        self.loss = evaluation_loss
        self.loss_names = evaluation_loss_names

        self.n_inputs = len(model.input_nodes)
        self.n_outputs = len(model.nodes_to_compute)
        self.n_targets = len(evaluation_loss.input_nodes) - self.n_outputs

        self.plot_maker = plot_maker
        self.db_info = db_info

        self.model_device = device_fallback()

    @property
    def var_list(self):
        return self.db_info["inputs"] + self.db_info["targets"]

    def evaluate(self, generator, eval_type, when=None):
        """
        Construct metrics and activate plotter on a dataloader.
        Note: the model is performed on whatever device it resides on.
        The validation loss is expected to be on the CPU -- this is not too slow,
        and allows for us to evaluate on large datasets whose results would take too much memory
        for the GPU.
        :param generator: the dataset loader
        :param eval_type: str describing dataset, such as 'train', 'valid', 'test'
        :param when: int (epoch #) or str describing when the evaluation is performed
        :return: loss_dict -- dictionary mapping loss names to loss values for this generator
        """

        # Storage for batch prediction values
        prediction_batch_vals = [[] for _ in range(self.n_outputs)]
        target_batch_vals = [[] for _ in range(self.n_targets)]

        # Get the batch prediction values
        for batch in progress_bar(generator, desc=f"Evaluating {eval_type}", unit="batch"):
            batch = [item.to(device=self.model_device) for item in batch]
            batch_inputs = batch[: self.n_inputs]
            batch_targets = batch[-self.n_targets :]
            batch_predictions = self.model(*batch_inputs)
            # Put predictions and targets on CPU
            for storage, value in zip(prediction_batch_vals, batch_predictions):
                storage.append(value.detach().cpu())

            for storage, value in zip(target_batch_vals, batch_targets):
                storage.append(value.detach().cpu())
            del batch_predictions #  To allow freeing memory.

        # Put the batches together
        prediction_all_vals = [torch.cat(x, dim=0) if x[0].shape != () else x[0] for x in prediction_batch_vals]
        target_all_vals = [torch.cat(x, dim=0) for x in target_batch_vals]

        # Evaluate the evaluation losses
        # Note: The `mean` here accounts for device-length vectors returned by scalars computed by
        # a DataParallel module.
        all_losses = [x.mean().item() for x in self.loss(*prediction_all_vals, *target_all_vals)]
        loss_dict = {name: value for name, value in zip(self.loss_names, all_losses)}

        if self.plot_maker:
            self.plot_maker.plot_phase(
                prediction_all_vals=prediction_all_vals, target_all_vals=target_all_vals, when=when, eval_type=eval_type
            )
        return loss_dict



"""
Functions for packing and unpacking tensor formats.
"""
from ..indextypes import register_index_transformer, IdxType

from ..nodes.indexers import QuadUnpackNode


@register_index_transformer(IdxType.QuadPack, IdxType.QuadMol)
def idx_QuadTriMol(node):
    parents = (node,)
    return parents, QuadUnpackNode



"""
Nodes for pair index state conversion
"""
from .. import find_unique_relative
from ..indextypes import register_index_transformer, IdxType, _debprint as idx_debprint
from ..nodes.base.node_functions import NodeOperationError
from ..nodes.indexers import PaddingIndexer, OneHotEncoder
from ..nodes.inputs import SpeciesNode, PositionsNode
from ..nodes.pairs import PairReIndexer, PairDeIndexer
from hippynn.graphs.nodes.tags import PairIndexer


# TODO: Rewrite so it can use non-one-hot-encodings?


@register_index_transformer(IdxType.MolAtomAtom, IdxType.Pair)
def idx_molatomatom_pair(node):
    purpose = "auto-generating indexing for {}".format(node)

    funrel = find_unique_relative  # Abbreviation because we so many calls in this function.
    if node.origin_node is None:
        # If we are auto-indexing in the model graph, it should be easy.
        pair_idx = funrel(node, PairIndexer, why_desc=purpose)
        pidxer = funrel(node, PaddingIndexer, why_desc=purpose)
        idx_debprint("Using reindexer for ", node)

    else:
        # If we are auto-indexing in the loss graph, it gets a bit complicated.

        # The species we link to will be the true version of species, that is, the node where
        species = funrel(node.origin_node, SpeciesNode, why_desc=purpose).true
        coordinates = funrel(node.origin_node, PositionsNode, why_desc=purpose).true
        try:
            encoder = funrel(species, OneHotEncoder, why_desc=purpose)
        except NodeOperationError as ne:
            idx_debprint("Creating new Padding indexer for {}".format(node))
            # If this fails, something bad has happened -- the loss graph is trying to do
            # something not defined by the model graph
            origin_encoder = funrel(species.origin_node, OneHotEncoder, why_desc=purpose)
            encoder = OneHotEncoder("one-hot", (species,), species_set=origin_encoder.species_set)
            # If we can't find an encoder, let's assume we won't find an indexer
            pidxer = PaddingIndexer("Auto(PaddingIndexer)", encoder)
        else:
            # If we did find an encoder, look for an indexer and assume it exists
            pidxer = funrel(encoder, PaddingIndexer, why_desc=purpose)

        try:
            pair_idx = funrel(species, PairIndexer, why_desc=purpose)
        except NodeOperationError as ne:
            idx_debprint("Creating new Pair indexer for {}".format(node))
            # If this fails, something bad has happened -- the loss graph is trying to do
            # something not defined by the model graph
            origin_pairs = funrel(species.origin_node, PairIndexer, why_desc=purpose)

            new_PairType = type(origin_pairs)

            pair_parents = (
                coordinates,
                pidxer.nonblank,
                pidxer.real_atoms,
                pidxer.inv_real_atoms,
            )
            try:
                cell = origin_pairs.cells
            except AttributeError:
                pass  # The pair indexer does not use cells
            else:
                pair_parents = pair_parents, *cell

            pair_idx = new_PairType("Pairs", pair_parents, dist_hard_max=origin_pairs.dist_hard_max)

    cls = PairReIndexer
    parents = node, pidxer.mol_index, pidxer.atom_index, pair_idx.pair_first, pair_idx.pair_second
    return parents, cls


@register_index_transformer(IdxType.Pair, IdxType.MolAtomAtom)
def idx_pair_molatomatom(node):
    purpose = "auto-generating indexing for {}".format(node)

    if node.origin_node is None:
        pidx = find_unique_relative(node, PairIndexer, why_desc=purpose)
        padidx = find_unique_relative(node, PaddingIndexer, why_desc=purpose)

        parents = (
            node,
            padidx.mol_index,
            padidx.atom_index,
            padidx.n_molecules,
            padidx.n_atoms_max,
            pidx.pair_first,
            pidx.pair_second,
        )

        return parents, PairDeIndexer
    else:
        raise NotImplementedError("De-indexing not yet implemented in loss graph")
        # TODO: refactor out padding indexer creation for loss from the molatom-atom indexer, then re-use it here.



import os
import warnings
import copy

from matplotlib import pyplot as plt

from ..graphs import GraphModule
from .. import settings


class PlotMaker:
    """
    The plot maker is responsible for collecting the data for the plotters and executing the plotters.
    """

    def __init__(self, *plotters, plot_every, save_dir="plots/"):
        self.plotters = plotters
        self.save_dir = save_dir
        self.plot_every = plot_every
        self.torch_module = None

    def assemble_module(self, model_outputs, targets):
        model_outputs = [x.pred for x in model_outputs]
        targets = [x.true for x in targets]
        all_inputs = (*model_outputs, *targets)
        graph = GraphModule(all_inputs, self.required_nodes)
        # Plot maker runs on CPU
        self.torch_module = copy.deepcopy(graph)
        self.torch_module.cpu()

    @property
    def required_nodes(self):
        return [p for plotter in self.plotters for p in plotter.parents]

    def make_plots(self, outputs, targets, sub_location=None):
        computed_data = self.torch_module(*outputs, *targets)
        remaining_data = computed_data

        location = self.make_full_location(sub_location)
        print("Making plots. Saved location: {}".format(location))

        for plotter in self.plotters:
            n_args = len(plotter.parents)
            this_data = remaining_data[:n_args]
            remaining_data = remaining_data[n_args:]
            if plotter.shown or plotter.saved:
                fig = plotter.make_plot(this_data)
                if plotter.shown:
                    plt.show()
                if plotter.saved:
                    file = os.path.join(location, plotter.saved)
                    print("Saving plot at {}".format(file))
                    fig.savefig(file, transparent=settings.TRANSPARENT_PLOT)
                plt.close(fig)

    def make_full_location(self, sub_location):
        sub_location = sub_location if sub_location else "Unspecified"
        location = os.path.join(self.save_dir, sub_location)
        if os.path.exists(location):
            warnings.warn("May override plots in {} !".format(location))
        else:
            os.makedirs(location)
        return location

    def plot_phase(self, prediction_all_vals, target_all_vals, when, eval_type):
        if when is not None:
            # If integer, convert to epoch number string IF it is time to plot.
            if isinstance(when, int):
                if (when % self.plot_every) == 0:
                    when = "epochs/epoch{}".format(when)
                else:
                    return
            # If string, make the plots. Note that this CANNOT become an ELIF, the above will convert when to a str.
            if isinstance(when, str):
                location = os.path.join(when, eval_type)
                self.make_plots(prediction_all_vals, target_all_vals, location)
            else:
                warnings.warn(
                    "Evaluation description was not a string "
                    + "or integer (specifying epoch number): {}\n".format(when)
                    + "Not generating plots."
                )



"""
Layers for pair finding
"""
import torch

from ...custom_kernels.utils import get_id_and_starts
from .open import _PairIndexer


class ExternalNeighbors(_PairIndexer):
    """
    module for ASE-like neighbors list fed into the graph
    """

    def forward(self, coordinates, real_atoms, shifts, cell, pair_first, pair_second):
        n_molecules, n_atoms, _ = coordinates.shape
        atom_coordinates = coordinates.reshape(n_molecules * n_atoms, 3)[real_atoms]
        paircoord = atom_coordinates[pair_second] - atom_coordinates[pair_first] + shifts.to(cell.dtype) @ cell
        distflat = paircoord.norm(dim=1)

        # Trim the list to only include relevant atoms, improving performance.
        within_cutoff_pairs = distflat < self.hard_dist_cutoff
        distflat = distflat[within_cutoff_pairs]
        pair_first = pair_first[within_cutoff_pairs]
        pair_second = pair_second[within_cutoff_pairs]
        paircoord = paircoord[within_cutoff_pairs]

        return distflat, pair_first, pair_second, paircoord


class PairReIndexer(torch.nn.Module):
    def forward(self, molatomatom_thing, molecule_index, atom_index, pair_first, pair_second):

        molecule_position = molecule_index[pair_first]
        absolute_first = atom_index[pair_first]
        absolute_second = atom_index[pair_second]
        out = molatomatom_thing[molecule_position, absolute_first, absolute_second]
        if out.ndimension() == 1:
            out = out.unsqueeze(1)
        return out


class PairDeIndexer(torch.nn.Module):
    def forward(self, features, molecule_index, atom_index, n_molecules, n_atoms_max, pair_first, pair_second):
        molecule_position = molecule_index[pair_first]
        absolute_first = atom_index[pair_first]
        absolute_second = atom_index[pair_second]

        if features.ndimension() == 1:
            features = features.unsqueeze(-1)
        featshape = features.shape[1:]
        out_shape = (n_molecules, n_atoms_max, n_atoms_max, *featshape)

        result = torch.zeros(*out_shape, device=features.device, dtype=features.dtype)
        result[molecule_position, absolute_first, absolute_second] = features
        return result


class MolPairSummer(torch.nn.Module):
    def forward(self, pairfeatures, mol_index, n_molecules, pair_first):
        pair_mol = mol_index[pair_first]
        if pairfeatures.shape[0] == 1:
            feat_shape = (1,)
            pairfeatures.unsqueeze(-1)
        else:
            feat_shape = pairfeatures.shape[1:]
        out_shape = (n_molecules, *feat_shape)
        result = torch.zeros(out_shape, device=pairfeatures.device, dtype=pairfeatures.dtype)
        result.index_add_(0, pair_mol, pairfeatures)
        return result


class PairCacher(torch.nn.Module):
    def __init__(self, n_images=1):
        super().__init__()
        self.set_images(n_images=n_images)

    def set_images(self, n_images):
        self.n_images = n_images

    def forward(
        self, pair_first, pair_second, cell_offsets, offset_index, real_atoms, mol_index, n_molecules, n_atoms_max
    ):
        # Set up absolute indices
        abs_atoms = real_atoms % n_atoms_max
        pfabs = abs_atoms[pair_first]
        psabs = abs_atoms[pair_second]
        mol = mol_index[pair_first]

        n_offsets = (2 * self.n_images + 1) ** 3

        # Reorder the indices
        order = offset_index + n_offsets * (psabs + n_atoms_max * (pfabs + n_atoms_max * mol))
        order = torch.argsort(order)
        mol = mol[order]
        pfabs = pfabs[order]
        psabs = psabs[order]
        offset_index = offset_index[order]
        cell_offsets = cell_offsets[order]

        # Create sparse tensor
        indices = torch.stack([mol, pfabs, psabs, offset_index], dim=0)
        values = cell_offsets
        size = (n_molecules, n_atoms_max, n_atoms_max, n_offsets, 3)
        s = torch.sparse_coo_tensor(
            indices=indices, values=values, size=size, dtype=torch.int, device=pair_first.device
        )
        s = s.coalesce()
        return s


class PairUncacher(torch.nn.Module):
    def __init__(self, n_images=1):
        super().__init__()
        self.set_images(n_images=n_images)

    def set_images(self, n_images):
        self.n_images = n_images

    def forward(self, sparse, coordinates, cell, real_atoms, inv_real_atoms, n_atoms_max, n_molecules):

        if not sparse.is_sparse:
            sparse = sparse.to_sparse(sparse_dim=4)
        sparse = sparse.coalesce()
        index = sparse.indices()
        values = sparse.values()

        mol, pfabs, psabs, offset_index = index.unbind(0)
        pfb = pfabs + mol * n_atoms_max
        psb = psabs + mol * n_atoms_max
        cell_offsets = values

        pair_first = inv_real_atoms[pfb]
        pair_second = inv_real_atoms[psb]

        atom_coordinates = coordinates.reshape(n_molecules * n_atoms_max, 3)[real_atoms]
        offsets = torch.bmm(cell_offsets.to(cell.dtype).unsqueeze(1), cell[mol]).squeeze(1)

        paircoord = atom_coordinates[pair_first] - atom_coordinates[pair_second] + offsets

        distflat = paircoord.norm(dim=1)

        return distflat, pair_first, pair_second, paircoord, cell_offsets, offset_index


def padded_neighlist(pair_first, pair_second, pair_coord, atom_array):
    """
    Convert from index list pair_first, pair_second
    to index list of `jlist`
    where jlist has shape (n_atoms, n_neigh_max).
    jlists consists of index of atom neighbors for each atom i,
    and is padded with values of -1.
    """

    # Scalar sizes required
    n_atoms = atom_array.shape[0]

    if pair_first.shape[0] == 0:  # empty neighbors list
        dev = pair_coord.device
        rijlist_pad = torch.empty((n_atoms, 0, 3), device=dev, dtype=pair_coord.dtype)
        jlist_pad = torch.empty((n_atoms, 0), device=dev, dtype=torch.int64)
        return jlist_pad, rijlist_pad

    with torch.no_grad():
        nneigh_max = torch.unique(pair_first, return_counts=True)[1].max()

        # Sorting pair list
        # Note, this isn't quite what resort_pairs does,
        # because resort_pairs only sorts on the first index
        # Note: The index is given so that the index functions
        # as a base-n_atoms representation of the pair of indices as a scalar number.
        # todo: possibly, just ensure a good sorting order for all pair finders.
        # todo: possibly, include cell offset information.
        # then this could be used to cache pairs without sparse tensors.
        ind = n_atoms * pair_first + pair_second
        sort = torch.argsort(ind)
        ilist = pair_first[sort]
        jlist = pair_second[sort]
        pair_coord = pair_coord[sort]

        #  List of where new i's start
        i_vals, istart = get_id_and_starts(ilist)
        istart = istart[:-1]

        # Relative index for j values
        # When boundaries of i-atoms are not crossed, we increment by 1
        diffrelj = torch.ones(len(jlist), dtype=jlist.dtype, device=jlist.device)
        # When boundaries of i-atoms are crossed, we decrement by the difference
        # in the i-atom position.
        diffrelj[istart[1:]] -= torch.diff(istart)
        # By summing the increments we get the index of j in the padded neighbors list
        jrellist = torch.cumsum(diffrelj, dim=0) - 1

        # Build Padded neighbor list and padded rij
        jlist_pad = -torch.ones(n_atoms, nneigh_max, dtype=ilist.dtype, device=ilist.device)
        jlist_pad[ilist, jrellist] = jlist

    rijlist_pad = torch.zeros(n_atoms, nneigh_max, 3, dtype=pair_coord.dtype, device=pair_coord.device)
    rijlist_pad[ilist, jrellist] = pair_coord

    return jlist_pad, rijlist_pad


class PaddedNeighModule(torch.nn.Module):
    def forward(self, pair_first, pair_second, pair_coord, atom_array):
        return padded_neighlist(pair_first, pair_second, pair_coord, atom_array)



"""
Test module for verifying implementation correctness against pytorch.
"""
import numba
import numpy as np
import torch

from . import env_pytorch
from . import autograd_wrapper
from . import env_numba


def get_simulated_data(
    n_molecules, n_atoms, atom_prob, n_features, n_nu, printinfo=False, dtype=None, device=torch.device("cpu")
):
    """
    Get semi-realistic test data for hipnn.
    n_molecules : number of molecules in the batch
    n_atoms     : number of atoms in each molecule
    atom_prob   : probability that an atom is real (not a padding)
    n_features  : number of features on e
    ach atom
    n_nu        : number of sensitivity functions

    Differences from real data:
        1)  Occupied atoms are (usually) first in the real arrays.
            This shouldn't matter because raw arrays are not output.

        2)  Sensitivity functions are randomly sparse with an
            average of 1 sensitivity non-zero per pair. Their amplitude is random.
            In real data, they are sparse, but 2 or 3 sequential ones will be on.
            Amplitudes have a concrete form based on distances
            Note: Sensitivities are not sparse here. There could be some use for
            representing them in a sparse way, though. I don't know how fast
            it would be to construct a sparse representation of sensitivities.

        3)  Sensitivity functions are not symmetric between pairs
            This necessitates summing over a pair (i,j) and (j, i) separately;
            certain recalculations are performed. I prefer code which works
            properly in the case of symmetric or non-symmetric sensitivities;
            we have some a case for asymmetric sensitivities (such as xACA),
            and in the future we may find use of nonsymmetric sensitivities.

        4)  Note that pairs never consist of the same atom repeatedly, that is,
            for all atoms i, (i,i) is not a pair.

    """
    if dtype is None:
        dtype = torch.get_default_dtype()
    np_fdtype = torch.zeros(1, dtype=torch.get_default_dtype()).numpy().dtype
    molatom_shp = (n_molecules, n_atoms)
    molatom_presence = np.random.choice([False, True], p=[1 - atom_prob, atom_prob], size=molatom_shp)
    atom_presence = molatom_presence.reshape(n_atoms * n_molecules)

    mol_features = np.random.randn(n_molecules, n_atoms, n_features).astype(np_fdtype)
    mol_features[~molatom_presence] = 0

    atom_features = mol_features.reshape(n_molecules * n_atoms, n_features).astype(np_fdtype)[atom_presence]
    n_real = molatom_presence.sum()
    real_atoms_arange = np.arange(n_real, dtype=int)
    inv_real_atoms = np.zeros(n_molecules * n_atoms, dtype=int)
    inv_real_atoms[atom_presence] = real_atoms_arange
    # atom_index = np.arange(n_molecules*n_atoms).reshape(molatom_shp)
    # mol_index = np.repeat(np.arange(n_molecules)[:,np.newaxis],n_atoms,axis=1)

    # pair_dists = np.sqrt(((coords[:,:,np.newaxis] - coords[:,np.newaxis,:])**2).sum(axis=3))
    pair_presence = (molatom_presence[:, np.newaxis, :] * molatom_presence[:, :, np.newaxis]) & (
        ~np.identity(n_atoms, dtype=bool)[np.newaxis, :, :]
    )

    atom_index = np.arange(n_molecules * n_atoms).reshape(molatom_shp)

    # We'll just do fully connected molecules.
    pair_first_pre = np.repeat(atom_index[:, :, np.newaxis], n_atoms, axis=2)[np.nonzero(pair_presence)]
    pair_second_pre = np.repeat(atom_index[:, np.newaxis, :], n_atoms, axis=1)[np.nonzero(pair_presence)]

    pair_first = inv_real_atoms[pair_first_pre]
    pair_second = inv_real_atoms[pair_second_pre]
    n_pairs = len(pair_first)

    # NOTE: These fake sensitivities are NONSYMMETRIC.
    # Current HIPNN does not do that, but a future one could.
    on_sensitivites = np.random.choice([True, False], p=[3 / n_nu, 1 - 3 / n_nu], size=(n_pairs, n_nu))
    pair_sensitivites = np.random.random(size=(n_pairs, n_nu)) * on_sensitivites
    assert not (pair_first == pair_second).any()
    if printinfo:
        print("Number of molecules:", n_molecules)
        print("Number of atoms per molecule:", n_atoms)
        print("Fraction of real atoms:", atom_prob)
        print("Number of features:", n_features)
        print("Number of sensitivities:", n_nu)
        print("Total atoms:", n_real)
        print("Total pairs:", n_pairs)
        # print("Total (any) atom pairs:",n_atoms**2*n_molecules)
        # print("Total (nonblank) atom pairs:",(molatom_presence[:,np.newaxis,:]*molatom_presence[:,:,np.newaxis]).sum())
    pair_sense = torch.tensor(pair_sensitivites).to(dtype=dtype, device=device)
    features = torch.tensor(atom_features).to(dtype=dtype, device=device)
    pair_first = torch.tensor(pair_first).to(device=device)
    pair_second = torch.tensor(pair_second).to(device=device)
    return pair_sense, features, pair_first, pair_second


TEST_TINY_PARAMS = dict(n_molecules=2, n_atoms=3, atom_prob=1., n_features=5, n_nu=7)
TEST_SMALL_PARAMS = dict(n_molecules=10, n_atoms=30, atom_prob=0.7, n_features=10, n_nu=20)
TEST_MEDIUM_PARAMS = dict(n_molecules=100, n_atoms=30, atom_prob=0.7, n_features=20, n_nu=20)
TEST_LARGE_PARAMS = dict(n_molecules=1000, n_atoms=30, atom_prob=0.7, n_features=80, n_nu=20)
TEST_MEGA_PARAMS = dict(n_molecules=500, n_atoms=30, atom_prob=0.7, n_features=128, n_nu=100)

# reference implementation


def test_pytorch_reference():
    sense, feat, pfirst, psecond = get_simulated_data(**TEST_TINY_PARAMS, dtype=torch.float64)
    sense.requires_grad_(True)
    feat.requires_grad_(True)
    env = env_pytorch.envsum(sense, feat, pfirst, psecond)
    pt_gradsense, pt_gradfeat = torch.autograd.grad(env.sum(), [sense, feat])
    sense.requires_grad_(False)
    feat.requires_grad_(False)
    ref_gradsense = env_pytorch.sensesum(torch.ones_like(env), feat, pfirst, psecond)
    ref_gradfeat = env_pytorch.featsum(torch.ones_like(env), sense, pfirst, psecond)

    assert torch.allclose(pt_gradsense, ref_gradsense)
    assert torch.allclose(pt_gradfeat, ref_gradfeat)
    assert (pt_gradsense == ref_gradsense).all()
    assert (pt_gradfeat == ref_gradfeat).all()


test_pytorch_reference()
del test_pytorch_reference


# A class for testing the correctness of the kernels functions.


class Envops_tester:
    def __init__(self, envsum_raw, sensesum_raw, featsum_raw, suspicious_deviation=0.5):
        self.envsum, self.sensesum, self.featsum, *_ = autograd_wrapper.wrap_envops(
            envsum_impl=envsum_raw, sensesum_impl=sensesum_raw, featsum_impl=featsum_raw
        )
        self.tol_f64 = dict(atol=1e-8, rtol=1e-5)
        self.tol_f32 = dict(atol=1e-5, rtol=1e-5)  # Absolute tolerance a bit fuzzier for float32.
        self.suspicious_deviation = suspicious_deviation

    def check_grad_and_gradgrad(self, func, differentiable_inputs, pair_first, pair_second, funcname=None):
        if funcname is None:
            funcname = func.__qualname__
        try:
            inputs = [x.requires_grad_(True) for x in differentiable_inputs]
            try:
                torch.autograd.gradcheck(func, [*inputs, pair_first, pair_second])
            except Exception as ee:
                raise ValueError(f"{funcname} failed grad check.") from ee
            try:
                torch.autograd.gradgradcheck(func, [*inputs, pair_first, pair_second])
            except Exception as ee:
                raise ValueError(f"{funcname} failed gradgrad check.") from ee
        finally:
            [x.requires_grad_(False) for x in inputs]

    def check_all_grad_once(self, device=torch.device("cpu")):
        sense, feat, pfirst, psecond = get_simulated_data(**TEST_TINY_PARAMS, dtype=torch.float64, device=device)
        n_atoms, n_features = feat.shape
        n_pairs, n_nu = sense.shape
        env = torch.rand(n_atoms, n_nu, n_features, dtype=sense.dtype, device=device)
        self.check_grad_and_gradgrad(self.envsum, (sense, feat), pfirst, psecond, funcname="envsum")
        self.check_grad_and_gradgrad(self.sensesum, (env, feat), pfirst, psecond, funcname="sensesum")
        self.check_grad_and_gradgrad(self.featsum, (env, sense), pfirst, psecond, funcname="featsum")

    def check_all_grad(self, repeats=3, device=torch.device("cpu")):
        for i in range(repeats):
            self.check_all_grad_once(device=device)

    def check_allclose_once(self, use_large=False, device=torch.device("cpu")):
        if use_large:
            sense, feat, pfirst, psecond = get_simulated_data(**TEST_LARGE_PARAMS, dtype=torch.float32, device=device)
        else:
            sense, feat, pfirst, psecond = get_simulated_data(**TEST_TINY_PARAMS, dtype=torch.float64, device=device)

        n_atoms, n_features = feat.shape
        n_pairs, n_nu = sense.shape
        env = torch.rand(n_atoms, n_nu, n_features, dtype=sense.dtype, device=device)
        for name, fn, pythfn, in1, in2 in (
            ("envsum", self.envsum, env_pytorch.envsum, sense, feat),
            ("sensesum", self.sensesum, env_pytorch.sensesum, env, feat),
            ("featsum", self.featsum, env_pytorch.featsum, env, sense),
        ):
            args = in1, in2, pfirst, psecond

            try:
                max_deviation = self.all_close_witherror(fn(*args), pythfn(*args))
            except Exception as ee:
                raise RuntimeError("Failed during {}".format(name)) from ee
            if max_deviation > self.suspicious_deviation:
                print("Closeness check for {} by suspicious amount".format(name), max_deviation)

    def check_empty(self,device=torch.device('cpu')):

        sense, feat, pfirst, psecond = get_simulated_data(**TEST_TINY_PARAMS, dtype=torch.float64, device=device)
        pfirst = psecond = torch.zeros((0,),dtype=torch.long,device=pfirst.device)    
        sense = torch.zeros((0,sense.shape[1]),dtype=sense.dtype,device=sense.device)
        
        try:
            env = self.envsum(sense,feat,pfirst,psecond)
            sense_g = self.sensesum(env,feat,pfirst,psecond)
            feat_g = self.featsum(env,sense,pfirst,psecond)
        except Exception as ee:
            raise ValueError("Failed an operation on data with zero pairs") from ee
        print("Passed zero-pair check")

    def all_close_witherror(self, r1, r2):
        r1 = r1.data.cpu().numpy()
        r2 = r2.data.cpu().numpy()
        if r1.dtype == np.float32:  # Get tolerance
            tol = self.tol_f32
        else:
            tol = self.tol_f64
        where = ~np.isclose(r1, r2, **tol)
        # np check formula: absolute(a - b) <= (atol + rtol * absolute(b))
        tol_arr = tol["atol"] + tol["rtol"] * np.abs(r2)
        diff_arr = np.abs(r1 - r2)
        violation = np.abs(diff_arr[where] / tol_arr[where])
        try:
            assert np.allclose(r1, r2, **tol)
        except AssertionError as aee:

            print("{} bad tolerances (of {})".format(np.sum(where), where.size))
            print("Allowed tolerance:")
            print(tol_arr[where])
            print("Observed deviations:")
            print(diff_arr[where])
            print("Ratio:")
            print(diff_arr[where] / tol_arr[where])
            print("Desired result:", r2[where])
            print("Observed result:", r1[where])
            print(r2.sum(), r1.sum())
            print(r1.sum(axis=1), r2.sum(axis=1))
            print(r1.sum(axis=0), r2.sum(axis=0))
            print("Locations:", np.nonzero(where))
            print("Violation stats: median: {}, max: {}".format(np.median(violation), np.max(violation)))
            raise aee
        return np.max(np.abs(diff_arr / tol_arr))  # max violation of bounds

    def check_allclose(self, repeats=30, use_large=False, device=torch.device("cpu")):
        for i in range(repeats):
            try:
                self.check_allclose_once(use_large, device=device)
            except Exception as ee:
                raise RuntimeError("Failed during iteration {}".format(i)) from ee

    def check_correctness(self, n_grad=1, n_small=100, n_large=3, device=torch.device("cpu")):
        self.check_empty(device=device)
        print("Checking gradients {} times...".format(n_grad))
        self.check_all_grad(repeats=n_grad, device=device)
        print("Passed gradient checks!")
        print("Checking forward methods on small data {} times...".format(n_small), flush=True)
        self.check_allclose(repeats=n_small, use_large=False, device=device)
        print("Passed small tensor forward checks!")
        print("Checking forward methods on large data {} times...".format(n_large), flush=True)
        self.check_allclose(repeats=n_large, use_large=True, device=device)
        print("Passed large tensor forward checks!")

    def check_speed(
        self, n_repetitions=10, device=torch.device("cpu"), data_size=TEST_LARGE_PARAMS, compare_against="Pytorch"
    ):

        if compare_against == "Pytorch":
            comp_envsum = env_pytorch.envsum
            comp_sensesum = env_pytorch.sensesum
            comp_featsum = env_pytorch.featsum
        elif compare_against == "Numba":
            comp_envsum = env_numba.new_envsum
            comp_sensesum = env_numba.new_sensesum
            comp_featsum = env_numba.new_featsum
        else:
            raise ValueError("Unknown implementation to comapre against:'{}'".format(compare_against))

        te, ts, tf = (TimerHolder(name) for name in ("Envsum", "Sensesum", "Featsum"))
        tne, tns, tnf = (
            TimerHolder("{}_{}".format(compare_against, name)) for name in ("Envsum", "Sensesum", "Featsum")
        )
        print("Repetitions: {}".format(n_repetitions))
        with torch.autograd.no_grad():
            # with torch.autograd.profiler.profile() as prof:
            for i in range(n_repetitions):
                print(".", end="", flush=True)
                sense, feat, pfirst, psecond = get_simulated_data(**data_size, dtype=torch.float32, device=device)
                with tne.add():
                    env = comp_envsum(sense, feat, pfirst, psecond)
                with tns.add():
                    comp_sensesum(env, feat, pfirst, psecond)
                with tnf.add():
                    comp_featsum(env, sense, pfirst, psecond)
                with te.add():
                    self.envsum(sense, feat, pfirst, psecond)
                with ts.add():
                    self.sensesum(env, feat, pfirst, psecond)
                with tf.add():
                    self.featsum(env, sense, pfirst, psecond)
        print()  # Newline to terminate the ... printing
        for t in [tne, tns, tnf] + [te, ts, tf]:
            print("Mean {} time: {} Median: {}".format(t.name, t.mean_elapsed, t.median_elapsed))
        for tn, t in zip([tne, tns, tnf], [te, ts, tf]):
            print("{} Speedup: {}".format(t.name, tn.median_elapsed / t.median_elapsed))

        tnsum = sum(x.median_elapsed for x in [tne, tns, tnf])
        tsum = sum(x.median_elapsed for x in [te, ts, tf])
        print("Overall {} time: {}".format(compare_against, tnsum))
        print("Overall time now: {}".format(tsum))
        print("Overall speedup: {}".format(tnsum / tsum))
        return  # prof


import time


class TimerHolder:
    def __init__(self, name=None):
        self.snippets = []
        self.name = name

    def add(self):
        t = TimedSnippet()
        self.snippets.append(t)
        return t

    @property
    def elapsed(self):
        return sum(t.elapsed for t in self.snippets)

    @property
    def mean_elapsed(self):
        return self.elapsed / len(self.snippets)

    @property
    def median_elapsed(self):
        return np.median([t.elapsed for t in self.snippets])


class TimedSnippet:
    def __init__(self):
        self.start = self.end = None

    def __enter__(self):
        if torch.cuda.is_available():
            torch.cuda.synchronize()
            numba.cuda.synchronize()
        self.start = time.time()

    def __exit__(self, exc_type, exc_value, exc_tb):
        if torch.cuda.is_available():
            torch.cuda.synchronize()
            numba.cuda.synchronize()
        self.end = time.time()

    @property
    def elapsed(self):
        return self.end - self.start


def main(env_impl,sense_impl,feat_impl):
    tester = Envops_tester(
        env_impl,
        sense_impl,
        feat_impl,
    )
    # % time

    if torch.cuda.is_available():
        print("Running GPU tests")
        meminfo = numba.cuda.current_context().get_memory_info()
        use_large_gpu = meminfo.free > 2 ** 31
        use_verylarge_gpu = meminfo.free > 2**3

        n_large = 3 if use_large_gpu else 0
        tester.check_correctness(device=torch.device("cuda"),n_large=n_large)
        if use_verylarge_gpu:
            print("-" * 80)
            print("Mega systems:", TEST_MEGA_PARAMS)
            tester.check_speed(n_repetitions=20,data_size=TEST_MEGA_PARAMS, device=torch.device("cuda"), compare_against="Pytorch")
        else:
            print("Numba indicates less than 32GB free GPU memory -- skipping mega system test")
        if use_large_gpu:
            print("-" * 80)
            tester.check_speed(n_repetitions=20,data_size=TEST_LARGE_PARAMS, device=torch.device("cuda"), compare_against="Pytorch") 
        else:
            print("Numba indicates less than 2GB free GPU memory -- skipping large system test")

        print("-" * 80)
        print("Medium systems:", TEST_MEDIUM_PARAMS)
        tester.check_speed(
            n_repetitions=100, data_size=TEST_MEDIUM_PARAMS, device=torch.device("cuda"), compare_against="Pytorch"
        )
        print("-" * 80)
        print("Small systems:", TEST_SMALL_PARAMS)
        tester.check_speed(
            n_repetitions=100, data_size=TEST_SMALL_PARAMS, device=torch.device("cuda"), compare_against="Pytorch"
        )

    else:
        print("Cuda not available, not running GPU tests.")

    print("Running CPU tests")
    tester.check_correctness()
    print("-" * 80)
    print("Large systems:", TEST_LARGE_PARAMS)
    tester.check_speed(n_repetitions=10, compare_against="Pytorch")
    print("-" * 80)
    print("Medium systems:", TEST_MEDIUM_PARAMS)
    tester.check_speed(n_repetitions=100, data_size=TEST_MEDIUM_PARAMS, compare_against="Pytorch")
    print("-" * 80)
    print("Small systems:", TEST_SMALL_PARAMS)
    tester.check_speed(n_repetitions=100, compare_against="Pytorch", data_size=TEST_SMALL_PARAMS)


if __name__ == "__main__":
    main(env_numba.new_envsum,
         env_numba.new_sensesum,
         env_numba.new_featsum,
    )



import torch

from scipy.spatial import KDTree

from .open import _PairIndexer, PairMemory
from torch.profiler import profile, record_function, ProfilerActivity

# Deprecated?
class StaticImagePeriodicPairIndexer(_PairIndexer):
    """Finds Pairs within a given number of images"""
    def __init__(self, *args, n_images=1, **kwargs):
        super().__init__(*args, **kwargs)
        import itertools

        # shape is (n_images+2)**3 by 3
        combinator = torch.tensor(list(itertools.product(range(-n_images, n_images + 1), repeat=3)))
        self.n_images = n_images
        self.register_buffer("combinator", combinator.to(torch.float))

    def forward(self, coordinates, nonblank, real_atoms, inv_real_atoms, cells):

        original_coordinates = coordinates

        with torch.no_grad():
            cur_device = coordinates.device

            # Figure out how many molecules and atoms we have
            n_molecules, n_atoms, _ = coordinates.shape
            n_atoms_max_batch = int(nonblank.sum(axis=1).max())
            # Construct a unique index for each atom (including blanks here)
            atom_index = torch.reshape(
                torch.arange(n_molecules * n_atoms, dtype=torch.long, device=cur_device), (n_molecules, n_atoms)
            )

            nonblank = nonblank[:, :n_atoms_max_batch]
            coordinates = coordinates[:, :n_atoms_max_batch]
            atom_index = atom_index[:, :n_atoms_max_batch]

            # Eliminate blank atom pairs, construct indices for them
            nonblank_pair = nonblank.unsqueeze(1) * nonblank.unsqueeze(2)
            nb_b, nb_a1, nb_a2 = torch.nonzero(nonblank_pair, as_tuple=True)

            # Get the indices of non-blank atoms
            pair_first = atom_index[nb_b, nb_a1]
            pair_second = atom_index[nb_b, nb_a2]
            # shape pair

            pair_diffcoords = coordinates[nb_b, nb_a1] - coordinates[nb_b, nb_a2]

            # Begin periodic part: find offsets for all images of cells.
            # Shape: n_shifts, n_batch, 3
            cell_shifts = torch.tensordot(self.combinator, cells, dims=((1,), (1,)))
            # Shape n_shifts,n_atoms,3
            offsets = cell_shifts[:, nb_b]

            # Find pairs/images close enough to be within the cutoff
            # shape n_shifts, pair
            pair_dist = (pair_diffcoords.unsqueeze(0) + offsets).norm(dim=2)
            close_pairs = (pair_dist < self.hard_dist_cutoff) & (pair_dist > 1e-5)  # get rid of self-pairs
            cp_offset, cp_pair = torch.nonzero(close_pairs, as_tuple=True)
            # cp_pair holds the indices of pairs
            # cp_offset holds the indices of the offsets

            # Extract the pairs within the cutoff
            pair_first = pair_first[cp_pair]
            pair_second = pair_second[cp_pair]
            pair_offsets = offsets[cp_offset, cp_pair]
            cell_offsets = self.combinator[cp_offset]

            # This converts the atom index from an index that counts blank atoms to one that doesn't;
            # Afterwards it indexes the flat array of atoms in the batch.
            pair_first = inv_real_atoms[pair_first]
            pair_second = inv_real_atoms[pair_second]

            del nb_b, nb_a1, nb_a2, cp_pair, pair_dist, offsets, pair_diffcoords, nonblank_pair, atom_index

        coordflat = original_coordinates.reshape(n_molecules * n_atoms, 3)[real_atoms]
        paircoord = coordflat[pair_first] - coordflat[pair_second] + pair_offsets
        distflat2 = paircoord.norm(dim=1)

        return distflat2, pair_first, pair_second, paircoord, cell_offsets, cp_offset


@torch.jit.script
def tracebatch(mat):
    """
    Trace of  of batch of matrices.
    Defaults to last two dimensions.
    """
    return torch.diagonal(mat, dim1=-2, dim2=-1).sum(dim=-1)

# For some reason torch.linalg on GPU tends to
# spend a lot of time allocating memory, especially
# when it is given a large batch of matrices.
# So we use the Cayley-Hamilton version of a matrix inverse
# Without calling any linalg functions.
@torch.jit.script
def cayley_inv(A):
    """
    Inverse of a batch of 3x3 matrices.
    """
    # Uses A^-1 = 1/det(A) ([tr(A)**2-tr(A**2)]I/2 - Atr(A) + A**2)
    assert A.shape[-1] == 3
    assert A.shape[-2] == 3

    trA = tracebatch(A)
    Asq = torch.bmm(A, A)
    trAsq = tracebatch(Asq)
    eye = torch.eye(3, dtype=A.dtype, device=A.device).broadcast_to(A.shape)
    eyeterm = (torch.pow(trA, 2) - trAsq).unsqueeze(-1).unsqueeze(-1) * eye / 2
    Ainvunscale = eyeterm - A * (trA.unsqueeze(-1).unsqueeze(-1)) + Asq
    det = tracebatch(torch.bmm(A, Ainvunscale)) / 3.

    return Ainvunscale / det.unsqueeze(-1).unsqueeze(-1)

@torch.jit.script
def wrap_systems_torch(coords, cell, cutoff: float):
    """
    Wraps coordinates to within the unit cell.

    Returns Inverse cells of shape (n_systems, 3 [cartesian], 3 [basis])
    Wrapped coords in same shape of coords
    Wrapped offset in shape (n_systems,n_atoms,3 [basis])
    and n_bounds in shape (n_systems,3) in terms of the number of image cells
    to search for each basis vector (radius).

    :param coords: coordinates of shape (n_systems,n_atoms,3)
    :param cell: cells of shape (n_systems, 3 [basis], 3 [cartesian])
    :param cutoff: cutoff radius to search.
    :return: inv_cells, wrapped_coords, wrapped_offset, n_bounds
    """
    # cell is (sys,basis,cartesian)

    # This is faster than calling the torch.linalg.inv directly as of torch 1.10
    inv_cell = cayley_inv(cell)
    # inv is (sys,cartesian,basis)
    projections = torch.bmm(coords, inv_cell)
    fractional_coords = torch.remainder(projections, 1)
    wrapped_offset = torch.div(projections, 1, rounding_mode="floor")
    wrapped_coords = torch.bmm(fractional_coords, cell)
    # wrapped_coords is (system,atoms,cartesian)

    # Compute number images to search in each basis direction
    eps = 1e-5
    inv_lengths = torch.linalg.norm(inv_cell, dim=1)
    n_bounds = (torch.floor(cutoff * inv_lengths + eps) + 1).to(torch.int64)

    return inv_cell, wrapped_coords, wrapped_offset.to(torch.int64), n_bounds


class PeriodicPairIndexer(_PairIndexer):
    """
    Finds pairs in general periodic conditions.
    """
    def forward(self, coordinates, nonblank, real_atoms, inv_real_atoms, cells):
        
        original_coordinates = coordinates

        with torch.no_grad():
            cur_device = coordinates.device

            # Figure out how many molecules and atoms we have
            n_molecules, n_atoms, _ = coordinates.shape
            n_atoms_max_batch = nonblank.sum(dim=1).max()
            # Construct a unique index for each atom (including blanks here)
            atom_index = torch.arange(n_molecules * n_atoms, dtype=torch.long, device=cur_device)
            atom_index = atom_index.reshape(n_molecules, n_atoms)

            # Trim padding in this batch to minimal level
            nonblank = nonblank[:, :n_atoms_max_batch]
            coordinates = coordinates[:, :n_atoms_max_batch]
            atom_index = atom_index[:, :n_atoms_max_batch]

            inv_cell, coordinates, wrapped_offsets, nbounds = (
                wrap_systems_torch(coordinates, cells, self.hard_dist_cutoff)
            )
            # inv cell: n_sys, n_cartesian, n_basis
            # wraped_coords: nsys, n_atoms, 3 (coords)
            # wrapped_offsets: nsys, n_atoms, 3 (basis)
            # nbounds: (nsys, 3) : number of cells-lengths to search in each direction.

            # ## make combinator of images ## #
            a1, a2, a3 = nbounds.max(dim=0)[0].unbind()
            a1 = torch.arange(-a1, a1+1, device=cur_device)
            a2 = torch.arange(-a2, a2+1, device=cur_device)
            a3 = torch.arange(-a3, a3+1, device=cur_device)
            combinator = torch.cartesian_prod(a1, a2, a3)
            # ## end make combinator ## #

            # Find images valid for each system
            inrange_images = (combinator.unsqueeze(1) >= -nbounds) & (combinator.unsqueeze(1) <= nbounds)
            inrange_images = inrange_images.all(dim=2).transpose(0, 1)
            # shape (n_sys, n_perm)

            # Eliminate blank atom pairs, construct indices for them
            nonblank_pair = nonblank.unsqueeze(1) * nonblank.unsqueeze(2)

            # ## Begin expensive part. ## #
            # In this zone we aggressively `del` things we don't need anymore
            # to allow pytorch to free memory. Delete anything that looks proportional
            # to the number of pairs. Many variables are proportional to the number of
            # 'possible' pairs, and all_sparse is proportional to an
            # even larger number of (# sys)*(# atoms)^2*(# possible images)

            # Start search space here.
            # Compute indices of all images for systems and nonblank atoms of those systems
            all_sparse = (inrange_images.unsqueeze(1).unsqueeze(2) * nonblank_pair.unsqueeze(3))
            nb_sys, nb_p1, nb_p2, nb_image = all_sparse.nonzero(as_tuple=True)
            del all_sparse

            # pair displacements without shifts
            pair_diffcoords = coordinates[nb_sys, nb_p1] - coordinates[nb_sys, nb_p2]

            # Find displacements offsets due to images.
            cell_shifts = torch.matmul(combinator.to(cells.dtype).unsqueeze(0), cells)
            # cell_shifts is  n_sys, n_perm, 3 (cartesian)
            dist_offsets = cell_shifts[nb_sys, nb_image]
            # Find pairs/images close enough to be within the cutoff
            pair_dist = (pair_diffcoords + dist_offsets).norm(dim=1)
            del dist_offsets, pair_diffcoords

            # Calculate mask of pairs that map to same atom.
            int_offsets = combinator[nb_image]
            nonself_pairs = torch.logical_or(int_offsets.to(torch.bool).any(dim=1), torch.ne(nb_p1, nb_p2))

            # Compute indices of considered pairs relative to MolAtom format.
            pair_first = atom_index[nb_sys, nb_p1]
            del nb_p1
            pair_second = atom_index[nb_sys, nb_p2]
            del nb_p2

            # Compute close pairs.
            close_pairs = (pair_dist < self.hard_dist_cutoff) & nonself_pairs  # get rid of self-pairs
            del nonself_pairs, pair_dist
            close_pairs = torch.nonzero(close_pairs)[:, 0]
            # close_pairs holds the indices of pairs

            # Extract the needed data within the cutoff.
            pair_mol = nb_sys[close_pairs]
            del nb_sys
            pair_first = pair_first[close_pairs]
            pair_second = pair_second[close_pairs]
            cell_only_offsets = int_offsets[close_pairs]  # Offset due to cell images only.
            offset_num = nb_image[close_pairs]  # index of where to store cell_offset values if caching pairs.
            del int_offsets, close_pairs
            # ## End expensive part ## #

            # This converts the atom index from an index that counts blank atoms (MolAtom) to one that doesn't;
            # Afterwards it indexes the flat array of atoms in the batch. (Atoms format)
            pair_first = inv_real_atoms[pair_first]
            pair_second = inv_real_atoms[pair_second]

            # Add back in offsets due to atom wrapping.
            atom_wrap_offsets = wrapped_offsets[nonblank]
            pf_shift = atom_wrap_offsets[pair_first]  # Offset of wrapping of first atom
            ps_shift = atom_wrap_offsets[pair_second]  # Offset of wrapping of second atom

            cell_offsets = cell_only_offsets - (pf_shift - ps_shift)  # Total image offset of pairs.
            del cell_only_offsets, pf_shift, ps_shift

        # Compute distance differentiably from total image offsets and pair indices
        pair_shifts = torch.matmul(cell_offsets.unsqueeze(1).to(cells.dtype), cells[pair_mol]).squeeze(1)
        coordflat = original_coordinates.reshape(n_molecules * n_atoms, 3)[real_atoms]
        paircoord = coordflat[pair_first] - coordflat[pair_second] + pair_shifts
        distflat2 = paircoord.norm(dim=1)

        return distflat2, pair_first, pair_second, paircoord, cell_offsets, offset_num, pair_mol
    
class PeriodicPairIndexerMemory(PairMemory):
    '''
    Implementation of PeriodicPairIndexer with additional memory component.

    Stores current pair indices in memory and reuses them to compute the pair distances if no 
    particle has moved more than skin/2 since last pair calculation. Otherwise uses the
    _pair_indexer_class to recompute the pairs.

    Increasing the value of 'skin' will increase the number of pair distances computed at
    each step, but decrease the number of times new pairs must be computed. Skin should be 
    set to zero while training for fastest results.
    '''
    _pair_indexer_class = PeriodicPairIndexer

    def forward(self, coordinates, nonblank, real_atoms, inv_real_atoms, cells):
        if self.recalculation_needed(coordinates, cells):
            self.n_molecules, self.n_atoms, _ = coordinates.shape
            self.recalculations += 1

            inputs = (coordinates, nonblank, real_atoms, inv_real_atoms, cells)
            outputs = self._pair_indexer(*inputs)
            distflat2, pair_first, pair_second, paircoord, cell_offsets, offset_num, pair_mol = outputs

            for name, var in [
                ("cell_offsets", cell_offsets),
                ("pair_first", pair_first),
                ("pair_second", pair_second),
                ("offset_num", offset_num),
                ("positions", coordinates),
                ("cells", cells),
                ("pair_mol", pair_mol)
                ]:
                self.__setattr__(name, var)

        else:
            self.reuses += 1
            pair_shifts = torch.matmul(self.cell_offsets.unsqueeze(1).to(cells.dtype), cells[self.pair_mol]).squeeze(1)
            coordflat = coordinates.reshape(self.n_molecules * self.n_atoms, 3)[real_atoms]
            paircoord = coordflat[self.pair_first] - coordflat[self.pair_second] + pair_shifts
            distflat2 = paircoord.norm(dim=1)

        return distflat2, self.pair_first, self.pair_second, paircoord, self.cell_offsets, self.offset_num


"""
Storage for the logical relationships between index states,
and a cache reduce redundant computations. Also includes
the indextransformer decorator which registers computations
that change index states.
The index transformers do not live here because they are themselves
nodes, and this package should be minimally dependent on the nodes themselves.
"""
import functools

from . import _debprint
from .type_def import IdxType

# fmt: off
_db_index_states = {
### TYPE PREDICTED BY MODEL | DEFAULT TYPE OF STORAGE IN DATABASE
    IdxType.Molecules       : IdxType.Molecules,
    IdxType.Atoms           : IdxType.MolAtom,
    IdxType.MolAtom         : IdxType.MolAtom,
    IdxType.MolAtomAtom     : IdxType.MolAtomAtom,
    IdxType.QuadMol         : IdxType.QuadPack,
    IdxType.QuadPack        : IdxType.QuadPack,
    IdxType.Pair            : IdxType.MolAtomAtom,
    IdxType.Scalar          : IdxType.Scalar,
    IdxType.NotFound        : IdxType.NotFound,
}


elementwise_compare_rules = {
#### TYPES OF ARGUMENTS TO REDUCE             |   TYPE OF OUTPUTS
    ((IdxType.Scalar,)                        ,   IdxType.Scalar),
    ((IdxType.Molecules,)                     ,   IdxType.Molecules),
    ((IdxType.Atoms,)                         ,   IdxType.Atoms),
    ((IdxType.MolAtom,)                       ,   IdxType.Atoms),
    ((IdxType.Atoms, IdxType.MolAtom,)        ,   IdxType.Atoms),
    ((IdxType.QuadMol,)                       ,   IdxType.QuadMol),
    ((IdxType.QuadPack,)                      ,   IdxType.QuadMol),
    ((IdxType.QuadMol, IdxType.QuadPack)      ,   IdxType.QuadMol),
    ((IdxType.Pair,)                          ,   IdxType.Pair),
    ((IdxType.Pair, IdxType.MolAtomAtom)      ,   IdxType.Pair),
    ((IdxType.MolAtomAtom,)                   ,   IdxType.Pair),
    ((IdxType.NotFound,)                      ,   IdxType.NotFound),
}
# fmt: on

# Add default rule: (_some_type,scalar) -> _some_type if scalar is not in rule already
for idxset, idxtarget in elementwise_compare_rules.copy():
    if IdxType.Scalar not in idxset:
        idxset = *idxset, IdxType.Scalar
        elementwise_compare_rules.add((idxset, idxtarget))
del idxset, idxtarget

# Assemble rules
elementwise_compare_rules = {frozenset(in_types): out_type for in_types, out_type in elementwise_compare_rules}

# The index cache holds information about previously indexed nodes.
# This information comes in two types
# 1) (parent, node_type, name) -> An generated index from an index transformer
# 2) (index_type, node) -> new_node a cache of `node` in index state `index_state`
_index_cache = {}

# The index dispatch table holds the transformer functions which are capable
# of finding the transformation between index states
_index_dispatch_table = {}


def clear_index_cache():
    """
    Remove any cached index types for nodes.

    The index cache holds global references to nodes that have been
    transformed automatically, so these nodes won't be garbage collected
    automatically. Clearing the cache will free any reference to them.

    If you modify the index dispatch table or type comparison table on the fly,
    you may want to clear the cache to remove references to nodes created
    using the old rules.
    """
    global _index_cache
    _index_cache = {}


def register_index_transformer(input_idxstate, output_idxstate):
    """
    Decorator for registering a transformer from one IdxType to another.

    An index transformer should have a signature::

         f(node) -> (parents, child_node_type)

    with types::

        node: Node
        parents : Tuple[Node]
        child_node_type : Union[Type[Node],Callable[[str,Tuple[node]],[node]]]

    That is, the child_node_type can be a class of node, or a factory function
    which acts with the same signature as a node constructor.
    If f supports additional arguments, they must have default values,
    as it will be invoked by the automatic index transformation system
    as f(node).

    The decorator results in a new function of type::

        f(node: Node) -> Node

    which is cached so that the same index transformation may be repeatedly
    applied to yield the same output node, and registered with the dispatch
    table for index transformations. No two functions may be registered for the
    same index type conversion simultaneously.

    """

    def decorator(f):
        if (input_idxstate, output_idxstate) in _index_dispatch_table:
            raise ValueError("Index dispatch for ({})->({}) already defined!".format(input_idxstate, output_idxstate))

        @functools.wraps(f)
        def wrapped(node, *args, **kwargs):
            idxcache_info = (output_idxstate, node)
            if idxcache_info in _index_cache:
                output_node = _index_cache[idxcache_info]
                _debprint("\tFound in node cache: {} : {}".format(output_node.name, output_node))
            else:
                parents, indexer_nodetype = f(node, *args, **kwargs)
                name = node.name + f"[{output_idxstate}]"
                output_node = indexer_nodetype(name, parents)

                # Forward-link this new node so we can avoid creating
                # redundant nodes.
                _index_cache[idxcache_info] = output_node
                # Back-link this new node to the old node's index state,
                # so if we end up with the reverse index transformation we
                # can just substitute the original node.
                _index_cache[(input_idxstate, output_node)] = node

                _debprint("\tGenerated new indexer: {}".format(output_node.name))
            return output_node

        _index_dispatch_table[input_idxstate, output_idxstate] = wrapped
        _debprint("Index dispatch for ({})->({}) defined!".format(input_idxstate, output_idxstate))
        return wrapped

    return decorator


#### Scalars can be viewed as any index type: Allow this ####
for idxt in IdxType:
    if idxt is IdxType.Scalar:
        continue  # No pass-through necessary for scalar-scalar

    # register no-op transformer
    @register_index_transformer(IdxType.Scalar, idxt)
    def scalar_promote(node, *, __idxt=idxt):
        _debprint("\t Allowing view of Scalar {} as type {}".format(node, __idxt))

        # Node index transformers (see #register_index_transformer) return
        # a set of parents and the node type to create.
        # So here we'll define a trivial factory function which mimics this
        # behavior and passes the node through unchanged.
        parents = node,  # Parents are defined as a tuple

        def passthrough_node_factory(name, parents):
            (node,) = parents  # Unwrap the tuple
            return node

        return parents, passthrough_node_factory

del scalar_promote, idxt



"""
Tools for organizing node definitions.

Usage of these tools is optional; they support node definitions
but are not required.

.. Note::
   The functions in this module are intended purely for defining new types
   of nodes. They are not strictly necessary; they only aid in defining more
   complex node behavior in a simple fashion.
"""
import functools
import contextlib
from .. import _debprint

from . import _BaseNode
from ...indextypes import index_type_coercion


class AutoNoKw:
    _auto_module_class = NotImplemented

    def auto_module(self):
        return self._auto_module_class()


class AutoKw:
    _auto_module_class = NotImplemented

    def auto_module(self):
        kw = self.module_kwargs or {}  # Default to empty dictionary if Falsey
        return self._auto_module_class(**kw)


@contextlib.contextmanager
def temporary_parents(child, parents):
    """
    Context manager for temporarily connecting a node to a set of parents.
    This is used during parent expansion so that `find_relatives` and
    `find_unique_relatives` can treat the nodes as connected even though
    they they are not fully formed.

    :param child:
    :param parents:
    :return: None
    """
    # Raise error if this is called on an already-build child.
    # (This function could be refactored to deal with this case.)
    assert not (
        hasattr(child, "parents") or hasattr(child, "children")
    ), "Temporary connection to node requires that it is not initialized."

    parset = set(parents)  # In case a node has the same parent twice.

    try:
        for p in parset:
            p.children = (*p.children, child)
        child.parents = parents
        child.children = ()
        yield
    finally:
        for p in parset:
            p.children = tuple(c for c in p.children if c is not child)
        del child.parents
        del child.children


class TupleTypeMismatch(Exception):
    pass


class AlwaysMatch:
    pass


def format_form_name(form):
    try:
        iter(form)
        formname = ", ".join(tuple(cls.__name__ for cls in form))
    except TypeError:
        formname = form.__name__
    return f"({formname})"


def _assert_tupleform(input_tuple, type_tuple):
    _debprint("ASSERTING FORM: ", input_tuple, type_tuple)

    # If it is a type of base node: return if it is the right type, or error.
    if isinstance(type_tuple, type):

        if type_tuple is AlwaysMatch:
            return True

        if isinstance(input_tuple, type_tuple):
            return True
        else:
            raise TupleTypeMismatch("Parent node did not match required type: {}".format(type_tuple))

    # If not, it must at least have the same length
    if not len(input_tuple) == len(type_tuple):
        raise TupleTypeMismatch(
            "Wrong length. {}!={}".format(len(input_tuple), len(type_tuple))
            + " \nInput: {} \nExpected: {}".format(input_tuple, type_tuple)
        )

    # If it does, it needs to have the correct type for each entry in the tuple.
    for i, (n, cls) in enumerate(zip(input_tuple, type_tuple)):
        if not isinstance(n, cls):
            raise TupleTypeMismatch(str(type_tuple) + ": wrong type at index {}".format(i))
    return True


# method wrapper for form storage
def adds_to_forms(fn):
    @functools.wraps(fn)
    def inner(self, *args, **kwargs):
        form = fn(self, *args, **kwargs)
        if self.matches is NotImplemented:
            self.matches = ()  # empty tuple
        self.matches = *self.matches, form

    return inner


class ParentExpander:
    """
    Manager object to register and implement
    optional steps in building a graph node.
    """

    def __init__(self):
        # Matches is a tuple of tuples of (form, matching function)
        self.matches = NotImplemented

    def __iter__(self):
        return iter(self.matches)

    def match(self, *form):
        """Decorator: the applied function will be a generic FormTransformer"""

        def inner(matchfn):
            if self.matches is NotImplemented:
                self.matches = ()  # empty tuple
            self.matches = *self.matches, FormTransformer(form, matchfn)
            return matchfn

        return inner

    def matchlen(self, length):
        """
        Decorator: The decorated function will be applied if the number
        of parents matches the given length.
        :param length:
        :return:
        """
        return self.match(*((_BaseNode,) * length))

    @adds_to_forms
    def assertion(self, *form):
        """
        Assert that the parents match a given form.

        :param form:
        :return:
        """
        return FormAssertion(form)

    @adds_to_forms
    def assertlen(self, length):
        """
        Assert that there are a given number of parents.

        :param length:
        :return:

        .. Note::
           It is recommended only to use this function once as the final stage
           of expanding a node's parents, to ensure that a node can be constructed
           directly from a satisfactory set of parents that doesn't require
           any expansion.
        """
        return FormAssertLength(length)

    @adds_to_forms
    def get_main_outputs(self):
        """
        Return the main outputs for any multinodes in the parents.

        :return:
        """
        return MainOutputTransformer(AlwaysMatch)

    @adds_to_forms
    def matched_idx_coercion(self, form, needed_index_states):
        """
        Apply coercion to the needed index states if the given form is present.
        :param form:
        :param needed_index_states:
        :return:
        """
        return IndexFormTransformer(form, needed_index_states)

    @adds_to_forms
    def require_idx_states(self, *needed_index_states):
        """
        Always coerce the nodes into a needed index state.

        :param needed_index_states:
        :return:

        .. Note::
           It is recommended only to use this function once as the final stage
           of expanding a node's parents, to ensure that a node can be constructed
           directly from a satisfactory set of parents that doesn't require
           any expansion.
        """
        return IndexFormTransformer(AlwaysMatch, needed_index_states)

    def _merge(self, *bases):
        """
        Used for merging

        :param bases: base classes to merge with (from MRO)
        :return: None
        """

        *true_bases, cls = reversed(bases)

        relevant_classes = tuple(
            sup_class
            for sup_class in true_bases
            if issubclass(sup_class, ExpandParents) and sup_class is not ExpandParents
        )

        base_matches = tuple(form for sup_class in relevant_classes for form in sup_class._parent_expander.matches)

        _debprint("Bases found for merging:", relevant_classes)

        if len(base_matches):
            # Filter only to contain the earliest copy of each form.
            if self.matches is not NotImplemented:
                base_matches = *base_matches, *self.matches
            base_form_set = set()
            base_match_list = []
            for match in base_matches:
                if match not in base_form_set:
                    _debprint("Adding", match)
                    base_form_set.add(match)
                    base_match_list.append(match)
                else:
                    _debprint("Already applied", match)
            base_matches = base_match_list
            _debprint("Forms found for merging:", base_matches)

            self.matches = tuple(base_matches)
        else:

            _debprint("No matches found for merging!")


class FormHandler:
    def add_class_doc(self):
        return None

    pass


def _fn_doc_adjust(form, fn):
    if fn.__module__ is __name__:
        # Don't document functions that live here, they are used many many times.
        return
    try:
        existing = fn.__doc__ or ""
        lines = existing.split("\n")
        # Count spaces in each non-blank line
        spaces = [len(l) - len(l.lstrip(" ")) for l in lines if l.strip() != ""]
        # Strip those with no spaces
        spaces = [x for x in spaces if x != 0]
        # Grab the min, if non exist, use 0
        spaces = min(spaces) if spaces else 0
        spaces = " " * spaces
        formname = format_form_name(form)
        fn.__doc__ = existing + f"\n{spaces}Used for creation from parents with signature {formname}\n{spaces}"

    except AttributeError:
        # Can't modify doc of a method dynamically.
        # This occurs when calling one of the explicit form handlers such as
        # AssertTypes or AssertLen
        pass


class FormTransformer(FormHandler):
    def __init__(self, form, fn):
        self.form = form
        self.fn = fn
        _fn_doc_adjust(form, fn)

    def add_class_doc(self):
        form_name = format_form_name(self.form)
        return f"If matching {form_name}, then apply " + (self.fn.__name__ or "")

    def __repr__(self):
        return f"ParentTransformer(sig={format_form_name(self.form)},fn={self.fn})"

    def __call__(self, node_self, *parents, purpose=None, **kwargs):
        try:
            _assert_tupleform(parents, self.form)
            _debprint("Expanding form", self.form)
            _debprint("Input parents:", self.form)
            _debprint("Match function:", self.fn)
            if purpose is None:
                purpose = "{}: Expanding parents {} based on form {}".format(type(self), parents, self.form)
            with temporary_parents(node_self, parents):
                # We have to pass self here explicitly because the form handler stores unbound functions.
                new_parents = self.fn(node_self, *parents, purpose=purpose, **kwargs)
            return new_parents
        except TupleTypeMismatch:
            _debprint("Didn't pass form!")
            return parents
        except Exception as ee:
            raise RuntimeError("Error while transforming {}".format(self.fn.__qualname__)) from ee


class IndexFormTransformer(FormTransformer):
    def __init__(self, form, idxstates):
        super().__init__(form, self.fn)
        self.idxstates = idxstates

    def add_class_doc(self):
        return f"Transforms the parents to have index states {self.idxstates}"

    def __repr__(self):
        return f"IndexTransformer from {format_form_name(self.form)} to {self.idxstates}"

    def fn(self, node_self, *parents, **kwargs):
        """Coerces index states for all parents"""
        new_parents = []
        for node, idxstate in zip(parents, self.idxstates):
            if idxstate is None:
                new = node
            else:
                new = index_type_coercion(node, idxstate)
            new_parents.append(new)
        return tuple(new_parents)


class MainOutputTransformer(FormTransformer):
    def __init__(self, form):
        super().__init__(form, self.fn)

    def add_class_doc(self):
        return """Gets main_output of nodes: casts MultiNodes to their main output"""

    @staticmethod
    def fn(node_self, *parents, **kwargs):
        """
        Gets main_output of nodes: casts multinodes to single nodes
        """
        return tuple(p.main_output for p in parents)


class FormAssertion(FormHandler):
    def __init__(self, form):
        self.form = form
        _debprint("Assertion created:", self)

    def add_class_doc(self):
        form_name = format_form_name(self.form)
        return f"Asserts that the parents have the form {form_name}."

    def __call__(self, self_cls, *parents, **kwargs):
        _debprint("Attempting Assertion", self)
        try:
            _assert_tupleform(parents, self.form)
        except TupleTypeMismatch as ee:
            raise TypeError("Input does not meet required form: \nInput: {} \nForm: {}".format(parents, self)) from ee
        return parents

    def __repr__(self):
        return super().__repr__() + "({})".format(self.form)


class FormAssertLength(FormAssertion):
    def __init__(self, length):
        self.length = length
        super().__init__((_BaseNode,) * length)

    def add_class_doc(self):
        return f"Asserts that the number of parents is {self.length}"

    def __repr__(self):
        return "FormLengthAssertion({})".format(self.length)


# This metaclass inserts the _parent_expander attribute into
# any class that has this metaclass.
# Note for developers: The metaclass is needed because
# the class must be modified before the class definition
# is executed; __init_subclass__ is run /after/ the class
# definition is executed.
class ExpandParentMeta(type):
    @classmethod
    def __prepare__(mcl, name, bases, **kwargs):
        cls_dict = super(ExpandParentMeta, mcl).__prepare__(name, bases, **kwargs)
        cls_dict["_parent_expander"] = ParentExpander()
        return cls_dict


def _append_docs(cls):
    """
    Add a note to documentation on parent expansion process.

    .. Note::
       This function assumes the class is defined at module-level scope.
       If not the indentation may be funny.
    """
    new_doc = (cls.__doc__ + "\n") if cls.__doc__ else ""
    new_doc += """\n    .. Note::\n       This node has parent expansion, following these procedures.\n\n"""
    for form_handler in cls._parent_expander:
        add = form_handler.add_class_doc()
        if add:
            new_doc += f"       #. {add}\n"
        else:
            raise ValueError("No documentation for transformation!")
    new_doc += "\n"
    return new_doc


class ExpandParents(metaclass=ExpandParentMeta):
    _parent_expander = None

    def __init_subclass__(cls, **kwargs):
        """
        Takes care of parent expansion setup and documentation for expansion of parents.
        """
        super().__init_subclass__(**kwargs)
        _debprint("Constructing parent expansion for ", cls)
        #  Note: Forms are applied in opposite order to MRO.
        #  This makes sense because initialization typically
        #  begins with the base class and then is customized or added to
        #  by a derived class.
        cls._parent_expander._merge(*cls.__mro__)

        # If no documentation, do not apply any decorating to the documentation
        if cls.__doc__:
            cls.__doc__ = _append_docs(cls)

    def expand_parents(self, parents, *, purpose=None, **kwargs):
        if isinstance(parents, _BaseNode):
            parents = (parents,)
        for form_handler in self._parent_expander:
            _debprint("Processing form:")
            _debprint("\tForm:", form_handler.form)
            _debprint("\t\targs:", parents)
            _debprint("\t\tkwargs:", kwargs)
            try:
                # We have to pass self here explicitly because the form handler stores unbound functions,
                # rather than bound methods.
                parents = form_handler(self, *parents, purpose=purpose, **kwargs)
            except Exception as ee:
                raise RuntimeError("Couldn't build {} automatically".format(type(self))) from ee
        return parents



"""
Layers that wrap other layers.
"""

import torch


# RESNET WRAPPER for the layers.
class ResNetWrapper(torch.nn.Module):
    """
    Resnet Wrapper Class
    """

    def __init__(self, base_layer, nf_in, nf_middle, nf_out, activation=torch.nn.functional.softplus):
        """
        Constructor

        :param base_layer: The ResLayer pytorch module
        :param nf_in: Input dimensions
        :param nf_out: Output dimensions
        :param activation: a nonlinearity function
        """
        super().__init__()

        self.activation = activation
        self.base_layer = base_layer
        self.res_layer = torch.nn.Linear(nf_middle, nf_out)

        if nf_in != nf_out:
            self.adjust_layer = torch.nn.Linear(nf_in, nf_out, False)
            self.needs_size_adjust = True
        else:
            self.needs_size_adjust = False

    def regularization_params(self):
        params = [self.res_layer.weight]
        if hasattr(self.base_layer, "regularization_params"):
            params.extend(self.base_layer.regularization_params())
        else:
            params.append(self.base_layer.weight)
        if self.needs_size_adjust:
            params.append(self.adjust_layer.weight)
        return params

    def forward(self, *input):
        """

        :param input: list of inputs for the layer;
            the first one is taken to be the features which will be computed as a residual
        :return: Pytorch module output
        """
        middle_activation = self.activation(self.base_layer(*input))
        difference_activation = self.res_layer(middle_activation)

        input_activation = input[0]
        if self.needs_size_adjust:
            input_activation = self.adjust_layer(input_activation)

        return difference_activation + input_activation



"""

This module is experimental.

:meta private:
"""
# Can't import this high-level node type before importing graphs.nodes,
# because it requires .graphs, and .graphs requires .nodes!
from collections import OrderedDict


# TODO
## COMPLETELY UNTESTED work in progress
# This was partially implemented as a way to think about mixed boundary conditions,
# so the graph processing can be conditional on some variable about the system
# Caveat Emptor: does not track modules used in the criteria functions!
# Criteria functions should have no learnable parameters!

import torch

from .base import SingleNode, MultiNode, AutoKw, ExpandParents


class _OptionModule(torch.nn.Module):
    def __init__(self, criteria_dict):
        super().__init__()

        # Similar indirection to GraphModule; see construction of GraphModule.moddict.
        self.criteria_names = OrderedDict((fn, "option{}".format(i)) for i, fn in enumerate(criteria_dict))
        self.moddict = torch.nn.ModuleDict({self.criteria_names[fn]: mod for fn, mod in self.criteria_dict.items()})

    def forward(self, option_input, *other_inputs):
        for fn, fn_name in zip(self.criteria_names):
            if fn(option_input):
                return self.moddict[fn_name](*other_inputs)
        raise RuntimeError("No criteria satisfied for option {}".format(option_input))


class OptionalMixin(ExpandParents, AutoKw):
    _auto_module_class = _OptionModule

    @_parent_expander.match(SingleNode)
    def _expand0(self, option, *, criteria_map, one_option):
        option_parents = one_option.parents
        assert all(
            o.parents == option_parents for k, o in criteria_map
        ), "All possible options must be linked to the same parents!"
        return (option, *option_parents)

    def __init__(self, name, parents, criteria_map, module="auto", **kwargs):
        one_option = criteria_map[list(criteria_map.keys())[0]]

        parents = self.expand_parents(parents, criteria_map, one_option)
        self.module_kwargs = {"criteria_map": {k: v.torch_module for k, v in criteria_map}}
        self._input_names = parents

        super().__init__(name, parents, module=module, **kwargs)


class OptionalNode(OptionalMixin, SingleNode):
    def __init__(self, name, parents, criteria_map, module="auto", **kwargs):
        assert all(isinstance(v, SingleNode) for k, v in criteria_map), "OptionalNode option types must be all BaseNode"
        one_option = one_value_from_dict(criteria_map)

        idxstate = one_option._index_state
        assert all(
            o._index_state == idxstate for o in criteria_map.values()
        ), "Sub-options must provide the same index structure to be compatible"
        self._index_state = idxstate

        super().__init__(self, name, parents, module=module, **kwargs)


class OptionalMultiNode(OptionalMixin, MultiNode):
    def __init__(self, name, parents, criteria_map, module="auto", **kwargs):
        one_option = one_value_from_dict(criteria_map)
        assert all(
            isinstance(v, SingleNode) for k, v in criteria_map
        ), "OptionalMultiNode option types must be all MultiNode"

        onames = one_option._output_names
        assert all(
            o._output_names == onames for o in criteria_map.values()
        ), "Sub-options must provide the same outputs to be compatible"
        self._output_names = onames

        child_states = one_option._output_index_states
        assert all(
            o._output_index_states == child_states for o in criteria_map.values()
        ), "Sub-options must provide the same index structure to be compatible"
        self._output_index_states = child_states

        super().__init__(self, name, parents, module=module, **kwargs)


def make_option(name, option_node, criteria_map, module="auto"):
    one_option = one_value_from_dict(criteria_map)
    if isinstance(one_option, MultiNode):
        return OptionalMultiNode(name, (option_node,), criteria_map, module=module)
    if isinstance(one_option, SingleNode):
        return OptionalNode(name, (option_node,), criteria_map, module=module)
    raise ValueError("No optional node constructor matching type: {}".format(type(one_option)))


def one_value_from_dict(d):
    return d[list(d.keys()[0])]



"""

Organized datasets for training and prediction.

.. Note::
   Databases constructed from disk (i.e. anything besides the base ``Database`` class)
   will load floating point data in the format (float32 or float64)
   specified via the ``torch.get_default_dtype()`` function. Use ``torch.set_default_dtype()``
   to control this behavior.

"""
from .database import Database
from .ondisk import DirectoryDatabase, NPZDatabase
has_ase = False
try: 
    import ase
    has_ase = True
except ImportError:
    pass

if has_ase:
   from ..interfaces.ase_interface import AseDatabase

all_list = ["Database", "DirectoryDatabase", "NPZDatabase"]
if has_ase:
    all_list += ["AseDatabase"]
__all__ = all_list



"""
Misc helpful functions which are not part of the library organization per se.
"""
import sys, os, traceback
import collections
import contextlib

import numpy as np
import torch

from . import settings


class teed_file_output:
    def __init__(self, *streams):
        self.streams = streams

    def write(self, *arg, **kwargs):
        for stream in self.streams:
            stream.write(*arg, **kwargs)

    def flush(self, *arg, **kwargs):
        for stream in self.streams:
            stream.flush(*arg, **kwargs)


@contextlib.contextmanager
def log_terminal(file, *args, **kwargs):
    """
    :param: file: filename or string
    :param: args:   piped to ``open(file,*args,**kwargs)`` if file is a string
    :param: kwargs: piped to ``open(file,*args,**kwargs)`` if file is a string

    Context manager where stdout and stderr are redirected to the specified file
    in addition to the usual stdout and stderr.
    The manager yields the file. Writes to the opened file object with
    "with log_terminal(...) as <file>" will not automatically
    be piped into the terminal.
    """
    if getattr(file, "write", None) is None:
        close_on_exit = True
        file = open(file, *args, **kwargs)
    else:
        close_on_exit = False
    teed_stderr = teed_file_output(file, sys.stderr)
    teed_stdout = teed_file_output(file, sys.stdout)
    with contextlib.redirect_stderr(teed_stderr):
        with contextlib.redirect_stdout(teed_stdout):
            try:
                yield file
            except:
                print("Uncaught exception, logging interrupted.", file=teed_stderr)
                traceback.print_exc(file=file)
                raise
            finally:
                if close_on_exit:
                    file.close()


@contextlib.contextmanager
def active_directory(dirname, create=None):
    """
    Context manager for temporarily switching the current working directory.

    If create is None, always succeed.
    If create is True, only succeed if the directory does not exist, and create one.
    If create is False, only succeed if the directory does exist, and switch to it.

    In other words, use create=True if you want to force that it's a new directory.
    Use create=False if you want to switch to an existing directory.
    Use create=None create a directory if you are okay with either alternative.

    :param dirname: directory to enter
    :param create: (None,True,False)

    :return: None
    :raises: If directory status not compatible with `create` constraints.
    """
    if create not in (None, True, False):
        raise ValueError("Argument 'create' must be one of None, True, or False")

    exists = os.path.exists(dirname)

    if create is True and exists:
        raise FileExistsError(
            f"Directory '{dirname}' exists already. Pass create=False or create=None to "
            "allow switching to an existing directory."
        )
    if create is False and not exists:
        raise FileNotFoundError(
            f"Directory '{dirname}' not found. Pass create=True or create=None to allow creation of the directory."
        )
    if not exists and create in (None, True):
        os.makedirs(dirname)

    return_dir = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(return_dir)


def progress_bar(iterable, *args, **kwargs):
    if settings.PROGRESS is None:
        return iterable
    else:
        return settings.PROGRESS(iterable, *args, **kwargs)


def param_print(module):
    count = 0
    for pname, p in module.named_parameters():
        count += p.numel()
        req_grad = "Learned" if p.requires_grad else "Fixed"
        print(p.device, req_grad, p.dtype, p.shape, pname)
    print("Total Count:", count)


def device_fallback():
    device = (torch.cuda.is_available() and torch.device(torch.cuda.current_device())) or torch.device("cpu")
    print("Device was not specified. Attempting to default to device:", device)
    return device


def arrdict_len(array_dictionary):
    """
    Return the length of one of the arrays in a dictionary. Under the assumption that they are all the same.
    :param array_dictionary:
    :return:
    """
    return len(next(iter(array_dictionary.values())))


def print_lr(optimizer):
    for i, param_group in enumerate(optimizer.param_groups):
        print("Learning rate:{:>10.5g}".format(param_group["lr"]))


def isiterable(obj):
    return isinstance(obj, collections.abc.Iterable)


def pad_np_array_to_length_with_zeros(array, length, axis=0):
    n = array.shape[axis]
    m = length - n
    if m < 0:
        raise ValueError(f"Cannot pad array to negative length! Array length: {n}, Total length requested: {length}")
    pad_width = [[0, 0] for _ in array.shape]
    pad_width[axis][1] = m
    return np.pad(array, pad_width, mode="constant")


def np_of_torchdefaultdtype():
    return torch.ones(1, dtype=torch.get_default_dtype()).numpy().dtype



"""
Things to do before training, i.e. initialization of network and diagnostics.
"""

import numpy as np
import torch

from .graphs import find_unique_relative, Predictor
from .graphs.nodes.base import _BaseNode
from .graphs.nodes.tags import Encoder
from .graphs.nodes.inputs import SpeciesNode, PositionsNode, CellNode, ForceNode
from .graphs.nodes.pairs import OpenPairIndexer, DynamicPeriodicPairs, MinDistNode
from .graphs.nodes.indexers import acquire_encoding_padding, SysMaxOfAtomsNode
from .graphs.nodes.physics import VecMag
from .networks.hipnn import compute_hipnn_e0


def set_e0_values(
    energy_module,
    database,
    trainable_after=False,
    decay_factor=1e-2,
    encoder=None,
    energy_name=None,
    species_name=None,
    peratom=False,
):
    """
    Computes values for the non-interacting energy using the training data.

    :param energy_module:   HEnergyNode or torch module for energy prediction
    :param database:        InterfaceDB object to get training data
    :param trainable_after: Determines if it should change .requires_grad attribute for the E0 parameters.
    :param decay_factor:    change initialized weights of further energy layers by ``df**N`` for layer N
    :param network_module:  network for running the species encoding. Can be auto-identified from energy node
    :param energy_name:     name for the energy variable, can be auto-identified from energy node
    :param species_name:    name for the species variable, can be auto-identified from energy node
    :return: None
    """

    if isinstance(energy_module, _BaseNode):
        if encoder is None:
            encoder = find_unique_relative(energy_module, Encoder, "Constructing E0 Values")
        if species_name is None:
            species_name = find_unique_relative(energy_module, SpeciesNode, "Constructing E0 Values").db_name
        if energy_name is None:
            energy_name = energy_module.main_output.db_name

        energy_module = energy_module.torch_module

    if isinstance(encoder, _BaseNode):
        encoder = encoder.torch_module

    train_data = database.splits["train"]

    z_vals = train_data[species_name]
    t_vals = train_data[energy_name]

    encoder.to(t_vals.device)
    eovals = compute_hipnn_e0(encoder, z_vals, t_vals, peratom=peratom)
    eo_layer = energy_module.layers[0]

    if not eo_layer.weight.data.shape[-1] == eovals.shape[-1]:
        raise NotImplementedError("The function set_eo_values does not currently work with custom InputNodes.")
    
    eo_layer.weight.data = eovals.reshape(1,-1)
    print("Computed E0 energies:", eovals)
    print("Computed E0 energies:", eovals)
    eo_layer.weight.data = eovals.expand_as(eo_layer.weight.data)
    print("Computed E0 energies:", eovals)   
    eo_layer.weight.data = eovals.expand_as(eo_layer.weight.data)
    
    eo_layer.weight.requires_grad_(trainable_after)
    for layer in energy_module.layers[1:]:
        layer.weight.data *= decay_factor
        layer.bias.data *= decay_factor
        decay_factor *= decay_factor


def _setup_min_dist_graph(
    species_name,
    positions_name,
    dist_hard_max,
    species_set,
    cell_name=None,
    pair_finder_class="auto",
    device=None,
):
    species = SpeciesNode(db_name=species_name)
    positions = PositionsNode(db_name=positions_name)

    # Species set required to build encoder and atom indexer.
    acquire_encoding_padding(species, species_set=species_set)

    if cell_name is not None:

        cell = CellNode(db_name=cell_name)

        if pair_finder_class == "auto":
            pair_finder_class = DynamicPeriodicPairs
        pair_parents = (positions, species, cell)

    else:
        if pair_finder_class == "auto":
            pair_finder_class = OpenPairIndexer
        pair_parents = (positions, species)

    pair_finder = pair_finder_class("PairFinder", pair_parents, dist_hard_max=dist_hard_max)

    min_dist_mol = MinDistNode("MinDists", pair_finder).min_dist_mol
    pred = Predictor(pair_parents, [min_dist_mol], model_device=device, name="Minimum Distance Calculator")

    return pred, min_dist_mol


def calculate_min_dists(
    array_dict: dict,
    species_name: str,
    positions_name: str,
    dist_hard_max: float,
    cell_name: str = None,
    device: torch.device = None,
    pair_finder_class: _BaseNode = "auto",
    batch_size: int = 50,
):
    """
    Calculates the minimum distance found in each system in ``array_dict``.

    Example usage for unsplit data::

    >>> db = Database(...)
    >>> min_dists = calculate_min_dists(db.arr_dict,"Z","R",5.0)

    If the database has been split::

    >>> min_dists_train = calculate_min_dists(db.splits['train'],"Z","R",5.0)

    Example usage to prune out low-distance data::

    >>> db = Database(...)
    >>> dist_threshold = ...
    >>> min_dist = calculate_min_dists(db.arr_dict,"Z","R",5.0)
    >>> low_distance_system = min_dist < dist_threshold
    >>> db.arr_dict = {k:v[~low_distance_system] for k,v in db.arr_dict.items()}

    .. Note::
       The cutoff radius ``dist_hard_max`` should be set large enough such that each atom is expected
       to have at least one neighbor. If an atom has no neighbors, its min_dist will be set to the
       largest distance found in the current *batch*.
       If an entire system has no neighbors, the minimum distance will be set to zero.

    :param array_dict: dictionary mapping strings to tensors/numpy arrays
    :param species_name: dictionary key for species
    :param positions_name: dictionary key for positions
    :param dist_hard_max: maximum distance to search
    :param cell_name: dictionary key for cell (periodic boundary conditions.
     if the cell is not specified, open boundaries are used.
    :param pair_finder_class: if 'auto', choose automatically. elsewise build this kind of pair finder.
    :param device: Where to perform the computation.
    :param batch_size: batch size to perform evaluation over.
    :return:
    """
    # Check for required info before proceeding to more expensive stuff.
    if species_name not in array_dict:
        raise KeyError(f"Species key {species_name} not in dictionary.")

    if positions_name not in array_dict:
        raise KeyError(f"Positions key {positions_name} not in dictionary.")

    if cell_name is not None:
        if cell_name not in array_dict:
            raise KeyError(f"Cell key {cell_name} not in dictionary.")

    species_set = list(np.unique(array_dict[species_name]))

    if 0 not in species_set:
        species_set = np.concatenate([[0], species_set], axis=0)

    if not len(species_set) or set(species_set) == {0}:
        raise ValueError("Species set empty!")

    pred, node_key = _setup_min_dist_graph(
        species_name=species_name,
        positions_name=positions_name,
        dist_hard_max=dist_hard_max,
        species_set=species_set,
        cell_name=cell_name,
        pair_finder_class=pair_finder_class,
        device=device,
    )

    input_dict = {
        species_name: torch.as_tensor(array_dict[species_name]),
        positions_name: torch.as_tensor(array_dict[positions_name]),
    }
    if cell_name is not None:
        input_dict[cell_name] = torch.as_tensor(array_dict[cell_name])

    results = pred(**input_dict, batch_size=batch_size)[node_key]

    return results


def _setup_max_force_graph(
    species_name,
    force_name,
    species_set,
    device=None,
):
    species = SpeciesNode(db_name=species_name)
    force = ForceNode(db_name=force_name)
    # Species set required to build encoder and atom indexer.
    enc, pad = acquire_encoding_padding(species, species_set=species_set)
    forcemag = VecMag("FMag", (force, species))
    max_force = SysMaxOfAtomsNode("MaxForce", (forcemag, pad))

    pred = Predictor([species, force], [max_force], model_device=device, name="Max Force Calculator")

    return pred, max_force


def calculate_max_system_force(
    array_dict: dict,
    species_name: str,
    force_name: str,
    device: torch.device = None,
    batch_size: int = 50,
):
    """
    Calculates the maximum force magnitude in each system in ``array_dict``.

    Example usage for unsplit data::

    >>> db = Database(...)
    >>> max_force = calculate_max_system_force(db.arr_dict,"Z","F")

    If the database has been split::

    >>> max_force_train = calculate_max_system_force(db.splits['train'],"Z","F")

    Example usage to prune out high-force data::

    >>> db = Database(...)
    >>> force_threshold = ...
    >>> max_force = calculate_max_system_force(db.arr_dict,"Z","F")
    >>> high_force_system = max_force > force_threshold
    >>> db.arr_dict = {k:v[~high_force_system] for k,v in db.arr_dict.items()}

    :param array_dict: dictionary mapping strings to tensors/numpy arrays
    :param species_name: dictionary key for species
    :param force-name: dictionary key for positions
    :param device: Where to perform the computation.
    :param batch_size: batch size to perform evaluation over.
    :return:
    """
    # Check for required info before proceeding to more expensive stuff.
    if species_name not in array_dict:
        raise KeyError(f"Species key {species_name} not in dictionary.")

    if force_name not in array_dict:
        raise KeyError(f"Positions key {force_name} not in dictionary.")

    species_set = list(np.unique(array_dict[species_name]))
    if 0 not in species_set:
        species_set = np.concatenate([[0], species_set], axis=0)

    if not len(species_set) or set(species_set) == {0}:
        raise ValueError("Species set empty!")

    pred, node_key = _setup_max_force_graph(
        species_name=species_name,
        force_name=force_name,
        species_set=species_set,
        device=device,
    )

    input_dict = {
        species_name: torch.as_tensor(array_dict[species_name]),
        force_name: torch.as_tensor(array_dict[force_name]),
    }

    results = pred(**input_dict, batch_size=batch_size)[node_key]

    return results



import torch
from seqm.basics import Energy, Force, parameterlist
from seqm.seqm_functions.constants import Constants
from seqm.seqm_functions.parameters import params
import os
from hippynn.graphs.nodes.base import MultiNode, AutoKw, find_unique_relative, ExpandParents, SingleNode
from hippynn.graphs.indextypes import IdxType
from hippynn.graphs.nodes.inputs import PositionsNode, SpeciesNode


class MLSEQM(torch.nn.Module):
    def __init__(self, seqm_parameters):
        super().__init__()
        self.learned = seqm_parameters["learned"]  # parameters for each real atom
        seqm_parameters["Hf_flag"] = False  # true: Hf, false: Etot-Eiso
        self.energy = Energy(seqm_parameters)
        self.const = Constants()
        self.elements = seqm_parameters["elements"]
        self.method = seqm_parameters["method"]
        self.filedir = seqm_parameters["parameter_file_dir"]
        self.n_output_parameters = len(self.learned)
        self.p = torch.nn.Parameter(
            params(method=self.method, elements=self.elements, root_dir=self.filedir, parameters=self.learned),
            requires_grad=True,
        )
        #
        self.const = Constants()
        self.save("./MOPAC/")  # save the initial value
        """
        #if only training to energy related terms:
        seqm_parameters['eig'] = False # don't need orbital energy
        seqm_parameters['scf_backward'] = 0  # ignore gradient on density matrix
        """

    def save(self, fdir="./MLSEQM/"):
        if not os.path.exists(fdir):
            os.system("mkdir %s" % fdir)
        fn = fdir + "parameters_" + self.method + "_MOPAC.csv"
        additional_parameters = [x for x in parameterlist[self.method] if x not in self.learned]
        p1 = params(method=self.method, elements=self.elements, root_dir=self.filedir, parameters=additional_parameters)
        with open(fn, "w") as f:
            f.write("N, " + ", ".join(["%23s" % s for s in self.learned]))
            if additional_parameters:
                f.write(", " + ", ".join(["%23s" % s for s in additional_parameters]))
            f.write("\n")
            for ele in self.elements:
                if ele == 0:
                    continue
                f.write(str(ele) + ", ")
                f.write(", ".join(["%23.16e" % x for x in self.p[ele]]))
                if additional_parameters:
                    f.write(", " + ", ".join(["%23.16e" % x for x in p1[ele]]))
                f.write("\n")

    def forward(self, coordinates, species):
        cond = species.reshape(-1) > 0
        Z = species.reshape(-1)[cond]
        learned_parameters = dict()
        for i in range(self.n_output_parameters):
            learned_parameters[self.learned[i]] = self.p[Z, i]

        Etot_m_Eiso, Etot, Eelec, Enuc, Eiso, EnucAB, e, P, charge, notconverged = self.energy(
            self.const, coordinates, species, learned_parameters=learned_parameters, all_terms=True
        )
        n_molecule, n_atom = species.shape
        atomic_charge = self.const.tore[species] - P.diagonal(dim1=1, dim2=2).reshape(n_molecule, n_atom, -1).sum(dim=2)

        return (
            Etot.reshape(-1, 1),
            Etot_m_Eiso.reshape(-1, 1),
            e,
            P,
            Eelec.reshape(-1, 1),
            Enuc.reshape(-1, 1),
            Eiso.reshape(-1, 1),
            charge,
            notconverged.reshape(-1, 1),
            atomic_charge,
        )


class MLSEQM_Node(AutoKw, MultiNode):
    _input_names = "Positions", "Species"
    _output_names = (
        "mol_energy",
        "Etot_m_Eiso",
        "orbital_energies",
        "single_particle_density_matrix",
        "electric_energy",
        "nuclear_energy",
        "isolated_atom_energy",
        "orbital_charges",
        "notconverged",
        "atomic_charge",
    )
    _main_output = "mol_energy"
    _output_index_states = (IdxType.Molecules,) * len(_output_names)
    _auto_module_class = MLSEQM

    def __init__(self, name, parents, seqm_parameters, module="auto", **kwargs):
        self.module_kwargs = dict(seqm_parameters=seqm_parameters)
        super().__init__(name, parents, module=module, **kwargs)



import seqm
import torch
from seqm.seqm_functions.fock import fock
from seqm.seqm_functions.hcore import hcore
from seqm.seqm_functions.diag import sym_eig_trunc1
from seqm.basics import Hamiltonian, Energy


class Hamiltonian_One(Hamiltonian):
    def __init__(self, seqm_parameters):
        super().__init__(seqm_parameters)

    def forward(
        self,
        const,
        molsize,
        nHeavy,
        nHydro,
        nOccMO,
        Z,
        maskd,
        mask,
        atom_molid,
        pair_molid,
        idxi,
        idxj,
        ni,
        nj,
        xij,
        rij,
        parameters,
        P0,
    ):
        beta = torch.cat((parameters["beta_s"].unsqueeze(1), parameters["beta_p"].unsqueeze(1)), dim=1)
        nmol = nOccMO.shape[0]
        zetas = parameters["zeta_s"]
        zetap = parameters["zeta_p"]
        uss = parameters["U_ss"]
        upp = parameters["U_pp"]
        gss = parameters["g_ss"]
        gsp = parameters["g_sp"]
        gpp = parameters["g_pp"]
        gp2 = parameters["g_p2"]
        hsp = parameters["h_sp"]
        M, w = hcore(
            const,
            nmol,
            molsize,
            maskd,
            mask,
            idxi,
            idxj,
            ni,
            nj,
            xij,
            rij,
            Z,
            zetas,
            zetap,
            uss,
            upp,
            gss,
            gpp,
            gp2,
            hsp,
            beta,
        )
        F = fock(nmol, molsize, P0, M, maskd, mask, idxi, idxj, w, gss, gpp, gsp, gp2, hsp)
        Hcore = M.reshape(nmol, molsize, molsize, 4, 4).transpose(2, 3).reshape(nmol, 4 * molsize, 4 * molsize)
        if self.eig:
            e, v = sym_eig_trunc1(F, nHeavy, nHydro, nOccMO, eig_only=True)
            return F, e, P0, Hcore, w, None, None
        else:
            return F, None, P0, Hcore, w, None, None


class Energy_One(Energy):
    def __init__(self, seqm_parameters):
        super().__init__(seqm_parameters)
        self.hamiltonian = Hamiltonian_One(seqm_parameters)


from hippynn.interfaces.pyseqm_interface.seqm_modules import SEQM_Energy, SEQM_All, pack_par


class SEQM_One_Energy(SEQM_Energy):
    def __init__(self, seqm_parameters):
        super().__init__(seqm_parameters)
        self.energy = Energy_One(seqm_parameters)

    def forward(self, par_atom, coordinates, species, single_particle_density_matrix):
        learned_parameters = pack_par(self, species, par_atom)

        Etot_m_Eiso, Etot = self.energy(
            self.const,
            coordinates,
            species,
            learned_parameters=learned_parameters,
            all_terms=True,
            P0=single_particle_density_matrix,
        )[:2]
        return Etot.reshape(-1, 1), Etot_m_Eiso.reshape(-1, 1)


class SEQM_One_All(SEQM_All):
    def __init__(self, seqm_parameters):
        super().__init__(seqm_parameters)
        self.energy = Energy_One(seqm_parameters)

    def forward(self, par_atom, coordinates, species, single_particle_density_matrix):
        learned_parameters = pack_par(self, species, par_atom)

        Etot_m_Eiso, Etot, Eelec, Enuc, Eiso, EnucAB, e, P, charge, notconverged = self.energy(
            self.const,
            coordinates,
            species,
            learned_parameters=learned_parameters,
            all_terms=True,
            P0=single_particle_density_matrix,
        )

        n_molecule, n_atom = species.shape
        atomic_charge = self.const.tore[species] - P.diagonal(dim1=1, dim2=2).reshape(n_molecule, n_atom, -1).sum(dim=2)

        return (
            Etot.reshape(-1, 1),
            Etot_m_Eiso.reshape(-1, 1),
            e,
            Eelec.reshape(-1, 1),
            Enuc.reshape(-1, 1),
            Eiso.reshape(-1, 1),
            atomic_charge,
        )


from hippynn.graphs.nodes.networks import Network
from hippynn.graphs.nodes.targets import HChargeNode  # , HBondNode
from hippynn.graphs.nodes.inputs import PositionsNode, SpeciesNode
from hippynn.graphs.nodes.base import InputNode
from hippynn.graphs.indextypes import IdxType
from hippynn.graphs.nodes.base import MultiNode, AutoKw, find_unique_relative, ExpandParents, SingleNode


class NotConvergedNode(InputNode):
    _index_state = IdxType.Molecules
    input_type_str = "notconverged"


class DensityMatrixNode(InputNode):
    _index_state = IdxType.Molecules
    input_type_str = "single_particle_density_matrix"


class SEQM_One_EnergyNode(ExpandParents, AutoKw, MultiNode):
    _input_names = "par_atom", "Positions", "Species", "single_particle_density_matrix"
    _output_names = "mol_energy", "Etot_m_Eiso"
    _main_output = "Etot_m_Eiso"
    _output_index_states = (IdxType.Molecules,) * len(_output_names)
    _auto_module_class = SEQM_One_Energy

    @_parent_expander.match(Network, DensityMatrixNode)
    def expand0(self, network, single_particle_density_matrix, seqm_parameters, decay_factor=1.0e-2, **kwargs):

        n_target_peratom = len(seqm_parameters["learned"])

        par_atom = HChargeNode(
            "SEQM_Atom_Params", network, module_kwargs=dict(n_target=n_target_peratom, first_is_interacting=True)
        )

        with torch.no_grad():
            for layer in par_atom.torch_module.layers:
                layer.weight.data *= decay_factor
                layer.bias.data *= decay_factor

        positions = find_unique_relative(network, PositionsNode)
        species = find_unique_relative(network, SpeciesNode)

        return par_atom.main_output, positions, species, single_particle_density_matrix

    _parent_expander.assertlen(4)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, None, None, None)

    def __init__(self, name, parents, seqm_parameters, decay_factor=1.0e-2, module="auto", **kwargs):
        parents = self.expand_parents(parents, seqm_parameters=seqm_parameters, decay_factor=decay_factor, **kwargs)
        self.module_kwargs = dict(seqm_parameters=seqm_parameters)
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_One_AllNode(SEQM_One_EnergyNode):
    _input_names = "par_atom", "Positions", "Species", "single_particle_density_matrix"
    _output_names = (
        "mol_energy",
        "Etot_m_Eiso",
        "orbital_energies",
        "electric_energy",
        "nuclear_energy",
        "isolated_atom_energy",
        "atomic_charge",
    )
    _main_output = "Etot_m_Eiso"
    _output_index_states = (IdxType.Molecules,) * len(_output_names)
    _auto_module_class = SEQM_One_All



import torch


class _PairIndexer(torch.nn.Module):
    def __init__(self, hard_dist_cutoff):
        super().__init__()
        self.hard_dist_cutoff = hard_dist_cutoff


class OpenPairIndexer(_PairIndexer):
    def forward(self, coordinates, nonblank, real_atoms, inv_real_atoms):

        with torch.no_grad():

            n_molecules, n_atoms, _ = coordinates.shape
            cur_device = coordinates.device
            # Figure out how many molecules and atoms we have

            atom_index = torch.reshape(
                torch.arange(n_molecules * n_atoms, dtype=torch.long, device=cur_device), (n_molecules, n_atoms)
            )
            n_atoms_max_batch = int(nonblank.sum(axis=1).max())
            namax_box = n_atoms_max_batch

            coordinates_trimmed = coordinates[:, :namax_box]
            atom_index = atom_index[:, :namax_box]
            nonblank_trimmed = nonblank[:, :namax_box]

            # Construct all the pair distances
            pair_dists = torch.norm(coordinates_trimmed.unsqueeze(2) - coordinates_trimmed.unsqueeze(1), dim=3)

            # Construct a unique index for each atom (including blanks here)

            # Pairs which are both not blank
            nonblank_pair = nonblank_trimmed.unsqueeze(1) * nonblank_trimmed.unsqueeze(2)
            # Don't connect an atom to itself
            nonself_atoms = (~torch.eye(namax_box, dtype=nonblank.dtype, device=cur_device)).unsqueeze(0)
            # Only take pairs that are close enough
            if self.hard_dist_cutoff is not None:
                close_atoms = pair_dists < self.hard_dist_cutoff
                pair_presence = nonblank_pair & nonself_atoms & close_atoms
            else:
                pair_presence = nonblank_pair & nonself_atoms

            # Now that we have a n_mol,n_atom,n_atom boolean mask of real pairs, we extract
            # the atom indices and pair distances associated with each.
            pair_indices = torch.nonzero(pair_presence, as_tuple=True)

            pair_first = atom_index.unsqueeze(2).expand(-1, -1, namax_box)[pair_indices]
            pair_second = atom_index.unsqueeze(1).expand(-1, namax_box, -1)[pair_indices]

            # This converts the pair index from an index that counts blank atoms to one that doesn't.
            pair_first = inv_real_atoms[pair_first]
            pair_second = inv_real_atoms[pair_second]

        coordflat = coordinates.reshape(n_molecules * n_atoms, 3)[real_atoms]
        paircoord = coordflat[pair_second] - coordflat[pair_first]
        distflat2 = paircoord.norm(dim=1)

        return distflat2, pair_first, pair_second, paircoord

class PairMemory(torch.nn.Module):
    '''
    Stores current pair indices and reuses them to compute the pair distances if no 
    particle has moved more than skin/2 since last pair calculation. Otherwise uses the
    _pair_indexer_class to recompute the pairs.

    Increasing the value of 'skin' will increase the number of pair distances computed at
    each step, but decrease the number of times new pairs must be computed. Skin should be 
    set to zero while training for fastest results.
    '''

    # ## Subclasses should update the following ## #
    _pair_indexer_class = NotImplemented

    def forward(*args, **kwargs):
        return NotImplementedError
    # ## End ## #

    def __init__(self, skin, dist_hard_max=None, hard_dist_cutoff=None):
        super().__init__()

        if dist_hard_max is None and hard_dist_cutoff is None:
            raise ValueError("One of 'dist_hard_max' and 'hard_dist_cutoff' must be specified.")
        if dist_hard_max is not None and hard_dist_cutoff is not None and dist_hard_max != hard_dist_cutoff:
            raise ValueError("Must only specify one of 'dist_hard_max' and 'hard_dist_cutoff.'")
        
        self.hard_dist_cutoff = (dist_hard_max or hard_dist_cutoff)
        self.dist_hard_max = (dist_hard_max or hard_dist_cutoff)
        self.set_skin(skin)

    @property
    def skin(self):
        return self._skin
    
    def set_skin(self, skin):
        self._skin = skin

        try:
            self._pair_indexer = self._pair_indexer_class(hard_dist_cutoff = self._skin + self.hard_dist_cutoff)
        except TypeError:
            self._pair_indexer = self._pair_indexer_class(dist_hard_max = self._skin + self.hard_dist_cutoff)

        self.reset_reuse_percentage()
        self.initialize_buffers()

    @skin.setter
    def skin(self, skin):
        self.set_skin(skin)
        
    @property
    def reuse_percentage(self):
        '''
        Returns None if there are no model calls on record.
        '''
        try:
            return self.reuses / (self.reuses + self.recalculations) * 100
        except ZeroDivisionError:
            print("No model calls on record.")
            return

    def reset_reuse_percentage(self):
        self.reuses = 0
        self.recalculations = 0
        
    def initialize_buffers(self):
        for name in ["pair_mol", "cell_offsets", "pair_first", "pair_second", "offset_num", "positions", "cells"]:
            self.register_buffer(name=name, tensor=None, persistent=False)

    def recalculation_needed(self, coordinates, cells):
        if coordinates.shape[0] != 1:  # does not support batch size larger than 1
            return True
        if self.positions is None:  # ie. forward function has not been called
            return True
        if self.skin == 0:
            return True
        if (self.cells != cells).any() or (((self.positions - coordinates)**2).sum(1).max() > (self._skin/2)**2):
            return True
        return False


import torch
from seqm.basics import Energy
from seqm.seqm_functions.constants import Constants
from seqm.seqm_functions.parameters import params
from seqm.seqm_functions.energy import elec_energy_isolated_atom

from .check import check, check_dist


class Scale(torch.nn.Module):
    def __init__(self, func=torch.sqrt):
        super().__init__()
        self.func = func

    def forward(self, notconverged):
        frac = 1.0 - notconverged.sum().double() / (notconverged.shape[0] + 0.0)
        return self.func(frac.detach())


class SEQM_MolMask(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, notconverged):
        mask = ~(notconverged.detach())
        return mask


class AtomMask(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, species):
        return species > 0


def num_orb(species, target_method="wB97X"):
    tore = Constants().tore.to(species.device)
    nHeavy = torch.sum(species > 1, dim=1)
    nHydro = torch.sum(species == 1, dim=1)

    # MNDO information
    nocc = (torch.sum(tore[species], dim=1) / 2.0).reshape(-1).type(torch.int64)
    norb = 4 * nHeavy + nHydro
    # nvirt = norb-nocc

    if target_method == "wB97X":
        # wB97X information
        # nocc0 = (torch.sum(species,dim=1)/2.0).type(torch.int64)
        # norb0 = torch.sum(species>0,dim=1)*4
        nvirt0 = torch.sum(species > 0, dim=1) * 4 - (torch.sum(species, dim=1) / 2.0).type(torch.int64)
        return torch.min(norb, nocc + nvirt0), nocc
    elif target_method == "gaussian":
        nvirt0 = (
            torch.sum(species > 1, dim=1) * 15
            + torch.sum(species == 1, dim=1) * 2
            - (torch.sum(species, dim=1) / 2.0).type(torch.int64)
        )
        return torch.min(norb, nocc + nvirt0), nocc
    else:
        raise KeyError(
            "Computing orbital number \
                        information is not implemented yet for ",
            target_method,
        )


"""
assume the target dataset is preprocessed such that
the data there has same number of occupied orbitals and at most the number of virtual orbitals
from SEQM.
as for certain molecules like O3, SEQM may have more virtual orbitals than from wB97X
For this case, the code here will trim the virtual orbitals from SEQM to match
"""


class SEQM_OrbitalMask(torch.nn.Module):
    def __init__(self, target_method, nOccVirt=None):
        """
        nOccVirt : if None, choose all the orbitals which MNDO and target methods share
                   else [NOCC, NVIRT], i.e. choose highest NOCC occupied orbtials (HOMO-NOCC+1, HOMO-NOCC+2, ..., HOMO)
                                            and lowest NVIRT virtual orbitals (LUMO, LUMO+1, ..., LUMO+NVIRT-1)
                        if there are not sufficient number of virtual/occupied orbitals, just choose whatever they have
                        i.e. NOCC <= nocc, NVIRT <= norb - nocc
        """
        super().__init__()
        self.target_method = target_method
        self.nOccVirt = nOccVirt

    def forward(self, species):
        norb, nocc = num_orb(species, self.target_method)
        nmol, molsize = species.shape
        tmp = (
            torch.arange(4 * molsize, dtype=torch.int64, device=species.device).reshape(1, -1).expand(nmol, 4 * molsize)
        )
        if isinstance(self.nOccVirt, (list, tuple)):
            NOCC, NVIRT = self.nOccVirt
            nindx1 = nocc - NOCC
            nindx1[nindx1 < 0] = 0
            nindx2 = nocc + NVIRT
            nindx2 = torch.min(norb, nindx2)

            mask = (tmp >= nindx1.reshape(-1, 1)) * (tmp < nindx2.reshape(-1, 1))
        else:
            mask = tmp < norb.reshape(-1, 1)

        return mask


class SEQM_MaskOnMol(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, var, mol_mask):
        if var.dim() == 1:
            return var.unsqueeze(1)[mol_mask.reshape(-1)]
        else:
            return var[mol_mask.reshape(-1)]


class SEQM_MaskOnMolAtom(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, var, mol_mask, atom_mask):
        # mol_mask shape : (mol, 1)
        # atom_mask shape : (mol, atom)
        mask = mol_mask * atom_mask
        return var[mask]


class SEQM_MaskOnMolOrbital(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, var, mol_mask, orbital_mask):
        # mol_mask shape : (mol, 1)
        # orbital_mask shape : (mol, orbital)
        # var shape : (mol, orbital)
        mask = mol_mask * orbital_mask
        return var[mask]


class SEQM_MaskOnMolOrbitalAtom(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, var, mol_mask, orbital_mask, atom_mask):
        # mol_mask shape : (mol, 1)
        # orbital_mask shape : (mol, orbital)
        # atom_mask shape : (mol, atom)
        # var shape : (mol, orbital, atom)
        mask = mol_mask.unsqueeze(2) * orbital_mask.unsqueeze(2) * atom_mask.unsqueeze(1)
        return var[mask]


def pack_par(obj, species, par_atom, par_bond=None):

    learned_parameters = {}
    eps = 1.0e-2
    if torch.is_tensor(par_bond):
        learned_parameters["Kbeta"] = par_bond + 1.0

    cond = species.reshape(-1) > 0
    Z = species.reshape(-1)[cond]

    for i in range(obj.n_output_parameters):
        # learned_parameters[obj.learned[i]] = par_atom[cond,i]*obj.weight[i]+obj.p[Z,i]
        learned_parameters[obj.learned[i]] = par_atom[:, i] * obj.weight[i] + obj.p[Z, i]

    for x in ["zeta_s", "zeta_p", "g_ss", "h_sp", "alpha", "Gaussian1_L", "Gaussian2_L", "Gaussian3_L", "Gaussian4_L"]:
        if x in obj.learned:
            learned_parameters[x] = obj.softplus(learned_parameters[x]) + eps

    assert ("g_pp" in obj.learned and "g_p2" in obj.learned) or (
        "g_pp" not in obj.learned and "g_p2" not in obj.learned
    ), "For convinience, both of g_pp and g_p2 should be provided from HIPNN or use constant ones"

    # hpp = 0.5*(gpp-gp2)>0
    if "g_pp" in obj.learned and "g_p2" in obj.learned:
        learned_parameters["g_pp"] = (
            obj.softplus(learned_parameters["g_pp"] - learned_parameters["g_p2"]) + learned_parameters["g_p2"] + eps
        )

    if obj.not_learned:
        # not_learned is a dict, e.g. : {1:['U_pp','h_sp']}, means use constant parameters for Hydrogen in U_pp and h_sp
        for k in obj.not_learned:
            for x in obj.not_learned[k]:
                if x in obj.learned:
                    i = obj.learned.index(x)
                    learned_parameters[x][Z == k] = obj.p[k, i]

    return learned_parameters


class SEQM_Energy(torch.nn.Module):
    def __init__(self, seqm_parameters):
        """
        Constructor
        """
        super().__init__()
        # ignore gradient on density matrix, based on Hellmann-Feynman Theorm
        # only works if targets are total energy or heat of formation
        # ?? for heat of formation, Hf = Etot - Eiso + Eexp
        # Eiso depends on parameters
        self.hartree_eV = 27.2113834
        self.learned = seqm_parameters["learned"]  # parameters for each real atom
        self.n_output_parameters = len(self.learned)
        seqm_parameters["eig"] = False  # don't need orbital energy
        seqm_parameters["scf_backward"] = 0  # ignore gradient on density matrix ``
        seqm_parameters["Hf_flag"] = False  # true: Hf, false: Etot-Eiso
        self.energy = Energy(seqm_parameters)
        self.const = Constants()
        self.elements = seqm_parameters["elements"]
        self.method = seqm_parameters["method"]
        self.filedir = seqm_parameters["parameter_file_dir"]
        self.p = params(method=self.method, elements=self.elements, root_dir=self.filedir, parameters=self.learned)
        self.softplus = torch.nn.Softplus(beta=5.0)
        self.weight = []
        for term in self.learned:
            if "zeta" in term or "alpha" in term or "Gaussian" in term:
                self.weight.append(1.0)
            else:
                self.weight.append(self.hartree_eV)
        self.weight = torch.nn.Parameter(torch.tensor(self.weight), requires_grad=False)
        self.not_learned = seqm_parameters.get("not_learned", None)

    def forward(self, par_atom, coordinates, species):
        """
        get the energy terms
        """
        learned_parameters = pack_par(self, species, par_atom)

        Etot_m_Eiso, Etot, Eelec, Enuc, Eiso, EnucAB, e, P, charge, notconverged = self.energy(
            self.const, coordinates, species, learned_parameters=learned_parameters, all_terms=True
        )

        return Etot.reshape(-1, 1), Etot_m_Eiso.reshape(-1, 1), notconverged.reshape(-1, 1)


class SEQM_All(torch.nn.Module):
    def __init__(self, seqm_parameters):
        """
        Constructor
        """
        super().__init__()
        self.hartree_eV = 27.2113834
        self.learned = seqm_parameters["learned"]  # parameters for each real atom
        self.n_output_parameters = len(self.learned)
        seqm_parameters["eig"] = True  # don't need orbital energy
        if "scf_backward" not in seqm_parameters:
            seqm_parameters["scf_backward"] = 1  # ignore gradient on density matrix ``
        elif seqm_parameters["scf_backward"] == 0:
            raise ValueError("scf_backward must be 1 or 2 for training with gradient on density matrix")
        seqm_parameters["Hf_flag"] = False  # true: Hf, false: Etot-Eiso
        self.energy = Energy(seqm_parameters)
        self.const = Constants()
        self.elements = seqm_parameters["elements"]
        self.method = seqm_parameters["method"]
        self.filedir = seqm_parameters["parameter_file_dir"]
        self.p = params(method=self.method, elements=self.elements, root_dir=self.filedir, parameters=self.learned)
        self.softplus = torch.nn.Softplus(beta=5.0)
        self.weight = []
        for term in self.learned:
            if "zeta" in term or "alpha" in term or "Gaussian" in term:
                self.weight.append(1.0)
            else:
                self.weight.append(self.hartree_eV)
        self.weight = torch.nn.Parameter(torch.tensor(self.weight), requires_grad=False)
        self.not_learned = seqm_parameters.get("not_learned", None)

    def forward(self, par_atom, coordinates, species):
        """
        get the energy terms
        """
        learned_parameters = pack_par(self, species, par_atom)

        Etot_m_Eiso, Etot, Eelec, Enuc, Eiso, EnucAB, e, P, charge, notconverged = self.energy(
            self.const, coordinates, species, learned_parameters=learned_parameters, all_terms=True
        )
        n_molecule, n_atom = species.shape
        atomic_charge = self.const.tore[species] - P.diagonal(dim1=1, dim2=2).reshape(n_molecule, n_atom, -1).sum(dim=2)

        return (
            Etot.reshape(-1, 1),
            Etot_m_Eiso.reshape(-1, 1),
            e,
            P,
            Eelec.reshape(-1, 1),
            Enuc.reshape(-1, 1),
            Eiso.reshape(-1, 1),
            charge,
            notconverged.reshape(-1, 1),
            atomic_charge,
        )



"""
For using hippynn with the Atomic Simulation Environment (ASE)

"""
from .calculator import HippynnCalculator, calculator_from_model

from .pairfinder import ASEPairNode

from .ase_database import AseDatabase

__all__ = ["HippynnCalculator", "calculator_from_model", "AseDatabase"]



"""
Nodes not otherwise categorized.
"""
from ..indextypes import IdxType
from .base import AutoNoKw, SingleNode, MultiNode
from ...layers import indexers as index_modules, algebra as algebra_modules


class StrainInducer(AutoNoKw, MultiNode):
    _input_names = "coordinates", "cell"
    _output_names = "strained_coordinates", "strained_cell", "strain"
    _output_index_states = NotImplemented
    _auto_module_class = index_modules.CellScaleInducer

    def __init__(self, name, parents, module="auto", **kwargs):
        position, cell = parents
        self._output_index_states = position._index_state, IdxType.NotFound, IdxType.NotFound
        super().__init__(name, parents, module=module, **kwargs)


class ListNode(AutoNoKw, SingleNode):
    _input_names = "features"
    _output_names = "wrapped_features"
    _auto_module_class = algebra_modules.ListMod

    def __init__(self, name, parents, module="auto"):
        super().__init__(name, parents, module=module)



"""
Keep track of training metrics over the experiment
"""
import copy
import warnings

try:
    from hippynn.plotting import plot_all_over_time
except ImportError:
    plot_all_over_time = lambda *args: warnings.warn(
        "Could not import plotting module. Make sure matplotlib is installed to enable plotting"
    )


class MetricTracker:
    """
    MetricTracker instances keep track of metrics and models for an experiment

    :ivar metric_names: the tracked metrics
    :ivar stopping_key: the metric used to determine what model is best
    :ivar best_metric_values: dictionary of metric keys to the best metric values observed so far.
          lower is assumed to be better.
    :ivar best_model: state dict for the best model so far.
    :ivar epoch_metric_values: list (index epoch) of
                               dictionary (key split, value of dictionary (key metric, value metric value)
    :ivar other_metric_values: dictionary of metric values at other times than after an epoch, for example
        the final test values.
    :ivar epoch_times: timing info for each epoch
    :ivar quiet: whether to print the values registered.

    """

    def __init__(self, metric_names, stopping_key, quiet=False, split_names=("train", "valid", "test")):
        """

        :param metric_names:
        :param stopping_key:
        :param quiet:
        :param split_names: splits to track.
        """
        if stopping_key not in metric_names and stopping_key is not None:
            raise ValueError("Stopping key {} is not in metric names {}".format(stopping_key, metric_names))
        self.metric_names = metric_names
        self.stopping_key = stopping_key

        # metadata
        self.name_column_width = max(len(i) for i in self.metric_names)  # name column size
        self.n_metrics = len(metric_names)

        # State variables
        self.best_metric_values = {split: {mtype: float("inf") for mtype in self.metric_names} for split in split_names}
        self.other_metric_values = {}
        self.best_model = None
        self.epoch_times = []

        self.epoch_metric_values = []
        self.epoch_best_metric_values = []
        self.quiet = quiet

    @classmethod
    def from_evaluator(cls, evaluator):
        return cls(evaluator.metric_names, evaluator.stopping_key)

    @property
    def current_epoch(self):
        return len(self.epoch_best_metric_values)

    def register_metrics(self, metric_info, when):
        """
        :param metric_info: dictionary of metric names: metric values
        :param when: string or integer specifying epoch number.
        :return:
        """
        """
        Updates the metrics for this epoch and computers which ones are better.
        
        """
        better_metrics = {k: {} for k in self.best_metric_values}
        for split_type, typevals in metric_info.items():
            for mname, mval in typevals.items():
                better = self.best_metric_values[split_type][mname] > mval
                if better:
                    self.best_metric_values[split_type][mname] = mval
                better_metrics[split_type][mname] = better
        if isinstance(when, int):
            self.epoch_metric_values.append(metric_info)
            self.epoch_best_metric_values.append(copy.deepcopy(self.best_metric_values))
        else:
            self.other_metric_values[when] = metric_info

        if self.stopping_key:
            better_model = better_metrics.get("valid", {}).get(self.stopping_key, False)
            stopping_key_metric = metric_info["valid"][self.stopping_key]
        else:
            better_model = None
            stopping_key_metric = None

        return better_metrics, better_model, stopping_key_metric

    def evaluation_print(self, evaluation_dict):
        if self.quiet:
            return
        table_evaluation_print(evaluation_dict, self.metric_names, self.name_column_width)

    def evaluation_print_better(self, evaluation_dict, better_dict):
        if self.quiet:
            return
        table_evaluation_print_better(evaluation_dict, better_dict, self.metric_names, self.name_column_width)
        if self.stopping_key:
            print(
                "Best {} so far: {:>8.5g}".format(
                    self.stopping_key, self.best_metric_values["valid"][self.stopping_key]
                )
            )

    def plot_over_time(self):
        plot_all_over_time(self.epoch_metric_values, self.epoch_best_metric_values)


# Driver for printing evaluation table results, with * for better entries.
# Decoupled from the estate in case we want to more easily change print formatting.
def table_evaluation_print_better(evaluation_dict, better_dict, metric_names, ncs):
    """
    Print metric results as a table, add a '*' character for metrics in better_dict.

    :param evaluation_dict: dict[eval type]->dict[metric]->value
    :param better_dict: dict[eval type]->dict[metric]->bool
    :param metric_names: Names
    :param ncs: Number of columns for name fields.
    :return: None
    """
    type_names = evaluation_dict.keys()
    better_labels = {True: "*", False: " "}

    transposed_values_better = [
        [(better_labels[better_dict[tname][mname]], evaluation_dict[tname][mname]) for tname in type_names]
        for mname in metric_names
    ]

    n_types = len(type_names)

    header = " " * (ncs + 2) + "".join("{:>14}".format(tn) for tn in type_names)
    rowstring = "{:<" + str(ncs) + "}: " + "   {}{:>10.5g}" * n_types

    print(header)
    print("-" * len(header))
    for n, valsbet in zip(metric_names, transposed_values_better):
        rowoutput = [k for bv in valsbet for k in bv]
        print(rowstring.format(n, *rowoutput))


# Driver for printing evaluation table results.
# Decoupled from the estate in case we want to more easily change print formatting.
def table_evaluation_print(evaluation_dict, metric_names, ncs):
    """
    Print metric results as a table.

    :param evaluation_dict: dict[eval type]->dict[metric]->value
    :param metric_names: Names
    :param ncs: Number of columns for name fields.
    :return: None
    """

    type_names, type_values = zip(*evaluation_dict.items())
    transposed_values = [[t[m] for t in type_values] for m in metric_names]

    n_types = len(type_names)

    header = " " * (ncs + 2) + "".join("{:>14}".format(tn) for tn in type_names)
    rowstring = "{:<" + str(ncs) + "}: " + "    {:>10.5g}" * n_types

    print(header)
    print("-" * len(header))
    for n, vals in zip(metric_names, transposed_values):
        print(rowstring.format(n, *vals))
    print("-" * len(header))



"""
Assembling graphs for a training run
"""
import copy
import warnings
import collections

from hippynn.graphs import GraphModule, get_subgraph, find_unique_relative
from hippynn.graphs.nodes.base import InputNode, LossInputNode, LossPredNode, LossTrueNode

from hippynn.experiment.evaluator import Evaluator

TrainingModules = collections.namedtuple("TrainingModules", ("model", "loss", "evaluator"))
"""
:param model: assembled torch.nn.Module of the model
:param loss: assembled torch.nn.Module for the loss
:param evaluator: assembled evaluator for validation losses
"""


def generate_database_info(inputs, targets, allow_unfound=False):
    """
    Construct db info from input nodes and target nodes.
    :param inputs: list of input nodes
    :param targets: list of target nodes
    :param allow_unfound: don't check if names are valid

    Builds a list of the db names for the nodes.
    If

    :return:
    """
    db_info = {
        "inputs": [i.db_name for i in inputs],
        "targets": [i.db_name for i in targets],
    }

    # Allows `None` to pass through.
    if allow_unfound:
        return db_info

    # If none of the names was `None`, return.
    if not any(name is None for row in db_info.values() for name in row):
        return db_info

    # Else, we need to raise an error.

    missing_inputs = [i for i, idb in zip(inputs, db_info["inputs"]) if idb is None]
    missing_targets = [i for i, idb in zip(targets, db_info["targets"]) if idb is None]

    msg = ""
    if missing_inputs:
        msg += "Missing inputs: {}\n".format(missing_inputs)
    if missing_targets:
        msg += "Missing targets: {}\n".format(missing_targets)

    raise ValueError("Required quantities for the graph inputs or targets are not mapped to a db name:\n{}".format(msg))


def build_loss_modules(training_loss, validation_losses, network_outputs, database_inputs):
    network_outputs = [x.pred for x in network_outputs]
    database_inputs = [x.true for x in database_inputs]
    all_inputs = (*network_outputs, *database_inputs)

    train_loss_module = GraphModule(all_inputs, [training_loss])
    valid_loss_module = GraphModule(all_inputs, validation_losses)
    valid_loss_module = copy.deepcopy(valid_loss_module)

    return train_loss_module, valid_loss_module


def determine_out_in_targ(*nodes_required_for_loss):
    """
    :param nodes_required_for_loss: train nodes and validation nodes
    :return: lists of needed network inputs, network outputs, and database targets
    """

    nodes_required_for_loss = get_subgraph(nodes_required_for_loss)

    required_inputs = {x for x in nodes_required_for_loss if isinstance(x, InputNode)}

    assert all(isinstance(x, LossInputNode) for x in required_inputs), (
        "Losses should not contain plain InputNode objects; only LossPredNode and LossTrueNode objects"
        "Bad nodes: {} ".format([x for x in required_inputs if not isinstance(x, LossInputNode)])
    )

    outputs = [x.origin_node for x in required_inputs if isinstance(x, LossPredNode)]
    targets = [x.origin_node for x in required_inputs if isinstance(x, LossTrueNode)]
    inputs = [p for node in outputs for p in node.get_all_parents() if isinstance(p, InputNode)]

    inputs = list(set(inputs))
    outputs = list(set(outputs))
    targets = list(set(targets))

    return inputs, outputs, targets


def assemble_for_training(train_loss, validation_losses, validation_names=None, plot_maker=None):
    """
    :param train_loss: LossNode
    :param validation_losses: dict of (name:loss_node) or list of LossNodes.
                              -if a list of loss nodes, the name of the node will be used for printing the loss,
                              -this can be overwritten with a list of validation_names
    :param validation_names (optional):  list of names for loss nodes, only if validation_losses is a list.
    :param plot_maker:        optional PlotMaker for model evaluation
    :return: training_modules, db_info
        -db_info: dict of inputs (input to model) and targets (input to loss) in terms of the `db_name`.

    ``assemble_for_training`` computes:

    #. what inputs are needed to the model
    #. what outputs of the model are needed for the loss
    #. what targets are needed from the database for the loss

    It then uses this info to create ``GraphModule`` s for the model and loss,
    and an ``Evaluator`` based on validation loss (& names), early stopping, plot maker.

    .. Note::
        Model and training loss are always evaluated on the active device.
        But the validation losses reside by default on the CPU. This helps compute statistics over large datasets.
        To accomplish this, the modules associated with the loss are copied in the validation loss.
        Thus, after assembling the modules for training, changes to the loss nodes will not affect the model evaluator.
        In all likelihood you aren't planning to do something too fancy like change the loss nodes during training.
        But if you do plan to do something like that with callbacks, know that you would probably need to construct a new
        evaluator.

    """
    if validation_names is None:
        if isinstance(validation_losses, dict):
            validation_names = list(validation_losses.keys())
            validation_losses = list(validation_losses.values())
        else:
            validation_names = [n.name for n in validation_losses]
    else:
        assert not isinstance(
            validation_losses, dict
        ), "Validation loss names cannot be supplied if validation_losses is a dictionary"

    if not all(isinstance(key, str) for key in validation_names):
        raise ValueError("Validation names must be strings.")

    loss_required_nodes = (train_loss, *validation_losses)

    if plot_maker is not None:
        loss_required_nodes = *loss_required_nodes, *plot_maker.required_nodes

    inputs, outputs, targets = determine_out_in_targ(*loss_required_nodes)

    print("Determined Inputs:", [x.name for x in inputs])
    print("Determined Outputs:", [x.name for x in outputs])
    print("Determined Targets:", [x.name for x in targets])

    model = GraphModule(inputs, outputs)

    loss_assembled, validation_lossfns = build_loss_modules(train_loss, validation_losses, outputs, targets)

    if plot_maker is not None:
        plot_maker.assemble_module(outputs, targets)

    db_info = generate_database_info(inputs, targets)

    evaluator = Evaluator(model, validation_lossfns, validation_names, plot_maker=plot_maker, db_info=db_info)

    return TrainingModules(model, loss_assembled, evaluator), db_info


_PAIRCACHE_DB_NAME = "AutoPrecomputedPairs"
from ..graphs.nodes import pairs, indexers, base
from ..graphs import Predictor, gops, inputs

_PAIRCACHE_COMPATIBLE_COMPUTERS = {pairs.NumpyDynamicPairs, pairs.PeriodicPairIndexer, pairs.DynamicPeriodicPairs}


def precompute_pairs(model, database, batch_size=10, device=None, make_dense=False, n_images=1):
    """

    :param model: Assembled GraphModule involving a PairIndexer
    :param database: database that precomputation should be supplied with
    :param batch_size: batch size to do pre-computation with
    :param device: where to do the precomputation.
    :param make_dense: return a dense array of pairs. Warning, this can be memory-expensive. However, it is necessary if you
        are going to use num_workers>0 in your dataloaders. If False, the cache is stored as a sparse array.
    :param n_images: number of images for cache storage, increase this if it fails.
        However, large values can incur a large memory cost if make_dense is True.

    :return: None-- changes the model graph.

    Note that the returned model needs to be re-assembled with the new graph for the cache to take effect.
    Example usage:
    >>> precompute_pairs(training_modules.model,database,device='cuda')
    >>> training_modules, db_info = assemble_for_training(train_loss, validation_losses)
    >>> database.inputs = db_info['inputs']
    """

    # nodes_to_compute,all_copies = gops.copy_subgraph(model.nodes_to_compute,assume_inputed=[])
    nodes_to_compute = model.nodes_to_compute
    pair_indexer = find_unique_relative(
        nodes_to_compute, lambda node: isinstance(node, tuple(_PAIRCACHE_COMPATIBLE_COMPUTERS))
    )
    dist_hard_max = pair_indexer.dist_hard_max
    cacher = pairs.PairCacher("PairCacher", pair_indexer, module_kwargs=dict(n_images=n_images))

    input_nodes = set([x for x in cacher.get_all_parents() if isinstance(x, base.InputNode)])
    pred = Predictor(input_nodes, [cacher], model_device=device, name="Pair Precomputer")

    try:
        for k, arrdict in database.splits.items():
            input_arrs = {node: arrdict[node.db_name] for node in input_nodes}

            outputs = pred(node_values=input_arrs, batch_size=batch_size)
            cache = outputs[cacher]
            if make_dense:
                cache = cache.to_dense()
            else:
                cache = cache.coalesce()
            database.splits[k][_PAIRCACHE_DB_NAME] = cache
    except RuntimeError as re:
        msg = re.args[0]
        if "size is inconsistent with indices" in msg:
            raise ValueError("Caching pairs required a larger number of images `n_images`")
        else:
            raise re

    cacheinput = inputs.PairIndices(db_name=_PAIRCACHE_DB_NAME)
    pos = gops.find_unique_relative(pair_indexer, inputs.PositionsNode)
    cell = gops.find_unique_relative(pair_indexer, inputs.CellNode)
    atomidx = gops.find_unique_relative(pair_indexer, indexers.AtomIndexer)
    uncacher = pairs.PairUncacher("AutoPairUncache",
                                  (cacheinput, pos, cell, atomidx),
                                  dist_hard_max=dist_hard_max)
    gops.replace_node(pair_indexer, uncacher, disconnect_old=True)

    return


# def reassemble(training_modules):
#     raise ValueError("Broken because of copying of loss graph for evaluator!")
#     model, loss_graph, evaluator=training_modules
#     train_loss = loss_graph.nodes_to_compute[0]
#     validation_losses = evaluator.loss.nodes_to_compute
#     validation_names = evaluator.loss_names
#     plot_maker = evaluator.plot_maker
#
#     return assemble_for_training(train_loss=train_loss,
#                                  validation_losses=validation_losses,
#                                  validation_names=validation_names,
#                                  plot_maker=plot_maker)



"""
Settings to use in the rest of the library.
"""

# Setup defaults.
# Note for developers: Settings should not affect
# the semantics of the models or training.
# The custom kernels may have very slight effects
# due to numerical differences.
import warnings
import os
import configparser

from distutils.util import strtobool
from types import SimpleNamespace
from functools import partial


try:
    from tqdm.contrib import tqdm_auto

    TQDM_PROGRESS = tqdm_auto
except ImportError:
    try:
        from tqdm import tqdm

        TQDM_PROGRESS = tqdm
    except ImportError:
        TQDM_PROGRESS = None

if TQDM_PROGRESS is not None:
    TQDM_PROGRESS = partial(TQDM_PROGRESS, mininterval=1.0, leave=False)

### Progress handlers


def progress_handler(prog_str):
    if prog_str == "tqdm":
        return TQDM_PROGRESS
    if prog_str.lower() == "none":
        return None
    warnings.warn(f"Unrecognized progress setting: '{prog_str}'. Setting to none.")


def kernel_handler(kernel_string):

    kernel_string = kernel_string.lower()

    kernel = {
        "0": False,
        "false": False,
        "pytorch": False,
        "1": True,
        "true": True,
    }.get(kernel_string, kernel_string)

    if kernel not in [True, False, "auto", "cupy", "numba"]:
        warnings.warn(f"Unrecognized custom kernel option: {kernel_string}. Setting custom kernels to 'auto'")
        kernel = "auto"

    return kernel


# keys: defaults, types, and handlers
default_settings = {
    "PROGRESS": (TQDM_PROGRESS, progress_handler),
    "DEFAULT_PLOT_FILETYPE": (".pdf", str),
    "TRANSPARENT_PLOT": (False, strtobool),
    "DEBUG_LOSS_BROADCAST": (False, strtobool),
    "DEBUG_GRAPH_EXECUTION": (False, strtobool),
    "DEBUG_NODE_CREATION": (False, strtobool),
    "DEBUG_AUTOINDEXING": (False, strtobool),
    "USE_CUSTOM_KERNELS": ("auto", kernel_handler),
    "WARN_LOW_DISTANCES": (True, strtobool),
}

settings = SimpleNamespace(**{k: default for k, (default, handler) in default_settings.items()})
settings.__doc__ = """
Values for the current hippynn settings.
See :doc:`/user_guide/settings` for a description.
"""

config_sources = {}  # Dictionary of configuration variable sources mapping to dictionary of configuration.
# We add to this dictionary in order of application

rc_name = os.path.expanduser("~/.hippynnrc")
if os.path.exists(rc_name) and os.path.isfile(rc_name):
    config = configparser.ConfigParser()
    config.read(rc_name)
    config_sources["~/.hippynnrc"] = config["GLOBALS"]

SETTING_PREFIX = "HIPPYNN_"
hippynn_environment_variables = {
    k.replace(SETTING_PREFIX, ""): v for k, v in os.environ.items() if k.startswith(SETTING_PREFIX)
}

LOCAL_RC_FILE_KEY = "LOCAL_RC_FILE"

if LOCAL_RC_FILE_KEY in hippynn_environment_variables:
    local_rc_fname = hippynn_environment_variables.pop(LOCAL_RC_FILE_KEY)
    if os.path.exists(local_rc_fname) and os.path.isfile(local_rc_fname):
        local_config = configparser.ConfigParser()
        local_config.read(local_rc_fname)
        config_sources[LOCAL_RC_FILE_KEY] = local_config["GLOBALS"]
    else:
        warnings.warn(f"Local configuration file {local_rc_fname} not found.")

config_sources["environment variables"] = hippynn_environment_variables

for sname, source in config_sources.items():
    for key, value in source.items():
        key = key.upper()
        if key in default_settings:
            default, handler = default_settings[key]
            try:
                setattr(settings, key, handler(value))
            except Exception as ee:
                raise ValueError(f"Value {value} for setting {key} is invalid") from ee
        else:
            warnings.warn(f"Configuration source {sname} contains invalid variables ({key}). They will not be used.")
